# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy
import os
import sys

import galsim
import galsim.des
from galsim_test_helpers import *
from galsim._pyfits import pyfits

@timer
def test_meds():
    """
    Create two objects, each with three exposures. Save them to a MEDS file.
    Load the MEDS file. Compare the created objects with the one read by MEDS.
    """
    # initialise empty MultiExposureObject list
    objlist = []

    # we will be using 2 objects for testing, each with 3 cutouts
    n_obj_test = 2
    n_cut_test = 3

    # set the image size
    box_size = 32

    # first obj
    img11 = galsim.Image(box_size, box_size, init_value=111)
    img12 = galsim.Image(box_size, box_size, init_value=112)
    img13 = galsim.Image(box_size, box_size, init_value=113)
    seg11 = galsim.Image(box_size, box_size, init_value=121)
    seg12 = galsim.Image(box_size, box_size, init_value=122)
    seg13 = galsim.Image(box_size, box_size, init_value=123)
    wth11 = galsim.Image(box_size, box_size, init_value=131)
    wth12 = galsim.Image(box_size, box_size, init_value=132)
    wth13 = galsim.Image(box_size, box_size, init_value=133)
    psf11 = galsim.Image(box_size, box_size, init_value=141)
    psf12 = galsim.Image(box_size, box_size, init_value=142)
    psf13 = galsim.Image(box_size, box_size, init_value=143)
    dudx = 11.1; dudy = 11.2; dvdx = 11.3; dvdy = 11.4; x0 = 11.5; y0 = 11.6;
    wcs11  = galsim.AffineTransform(dudx, dudy, dvdx, dvdy, galsim.PositionD(x0, y0))
    dudx = 12.1; dudy = 12.2; dvdx = 12.3; dvdy = 12.4;
    wcs12  = galsim.JacobianWCS(dudx, dudy, dvdx, dvdy)
    wcs13  = galsim.PixelScale(13)


    # create lists
    images = [img11, img12, img13]
    weight = [wth11, wth12, wth13]
    seg =    [seg11, seg12, seg13]
    psf =    [psf11, psf12, psf13]
    wcs =    [wcs11, wcs12, wcs13]

    # create object
    obj1 = galsim.des.MultiExposureObject(images=images, weight=weight, seg=seg, psf=psf,
                                          wcs=wcs, id=1)

    # second obj
    img21 = galsim.Image(box_size, box_size, init_value=211)
    img22 = galsim.Image(box_size, box_size, init_value=212)
    img23 = galsim.Image(box_size, box_size, init_value=213)
    seg21 = galsim.Image(box_size, box_size, init_value=221)
    seg22 = galsim.Image(box_size, box_size, init_value=222)
    seg23 = galsim.Image(box_size, box_size, init_value=223)
    wth21 = galsim.Image(box_size, box_size, init_value=231)
    wth22 = galsim.Image(box_size, box_size, init_value=332)
    wth23 = galsim.Image(box_size, box_size, init_value=333)
    psf21 = galsim.Image(box_size, box_size, init_value=241)
    psf22 = galsim.Image(box_size, box_size, init_value=342)
    psf23 = galsim.Image(box_size, box_size, init_value=343)

    dudx = 21.1; dudy = 21.2; dvdx = 21.3; dvdy = 21.4; x0 = 21.5; y0 = 21.6;
    wcs21  = galsim.AffineTransform(dudx, dudy, dvdx, dvdy, galsim.PositionD(x0, y0))
    dudx = 22.1; dudy = 22.2; dvdx = 22.3; dvdy = 22.4;
    wcs22  = galsim.JacobianWCS(dudx, dudy, dvdx, dvdy)
    wcs23  = galsim.PixelScale(23)

    # create lists
    images = [img21, img22, img23]
    weight = [wth21, wth22, wth23]
    seg =    [seg21, seg22, seg23]
    psf =    [psf21, psf22, psf23]
    wcs =    [wcs21, wcs22, wcs23]

    # create object
    # This time put the wcs in the image and get it there.
    img21.wcs = wcs21
    img22.wcs = wcs22
    img23.wcs = wcs23
    obj2 = galsim.des.MultiExposureObject(images=images, weight=weight, seg=seg, psf=psf, id=2)

    obj3 = galsim.des.MultiExposureObject(images=images, id=3)

    # create an object list
    objlist = [obj1, obj2]

    # save objects to MEDS file
    filename_meds = 'output/test_meds.fits'
    galsim.des.WriteMEDS(objlist, filename_meds, clobber=True)

    bad1 = galsim.Image(32, 48, init_value=0)
    bad2 = galsim.Image(35, 35, init_value=0)
    bad3 = galsim.Image(48, 48, init_value=0)

    with assert_raises(TypeError):
        galsim.des.MultiExposureObject(images=img11)
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[])
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[bad1])
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[bad2])
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[img11,bad3])
    with assert_raises(TypeError):
        galsim.des.MultiExposureObject(images=images, weight=wth11)
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=images, weight=[])
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[img11], weight=[bad3])
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[img11], psf=[bad1])
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[img11], psf=[bad2])
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[img11, img12], psf=[bad2, psf12])
    with assert_raises(TypeError):
        galsim.des.MultiExposureObject(images=images, wcs=wcs11)
    celestial_wcs = galsim.FitsWCS("DECam_00154912_12_header.fits", dir='des_data')
    with assert_raises(galsim.GalSimValueError):
        galsim.des.MultiExposureObject(images=[img11], wcs=[celestial_wcs])

    # Check the one with no psf, weight, etc.
    filename_meds2 = 'output/test_meds_image_only.fits'
    galsim.des.WriteMEDS([obj3], filename_meds2, clobber=True)


    # Note that while there are no tests prior to this, the above still checks for
    # syntax errors in the meds creation software, so it's still worth running as part
    # of the normal unit tests.
    # But for the rest of the tests, we'll use the meds module to make sure our code
    # stays in sync with any changes there.
    try:
        import meds
        # Meds will import this, so check for this too.
        import fitsio
    except ImportError:
        print('Failed to import either meds or fitsio.  Unable to do tests of meds file.')
        return

    # Run meds module's validate function
    try:
        meds.util.validate_meds(filename_meds)
        meds.util.validate_meds(filename_meds2)
    except AttributeError:
        print('Seems to be the wrong meds package.  Unable to do tests of meds file.')
        return

    m = meds.MEDS(filename_meds)

    # Check the image_info extension:
    ref_info = meds.util.get_image_info_dtype(1)
    info = m.get_image_info()
    print('info = ',info)
    for name, dt in ref_info:
        dt = numpy.dtype(dt)
        print(name, dt, info.dtype[name], dt.char, info.dtype[name].char)
        assert name in info.dtype.names, "column %s not present in image_info extension"%name
        # I think S and U for this purpose are equivalent.
        # But I'm finding S in the reference, and U in info.
        c = info.dtype[name].char
        c = 'S' if c == 'U' else c
        assert dt.char == c, "column %s is the wrong type"%name

    # Check the basic structure of the object_data extension
    cat = m.get_cat()
    ref_data = meds.util.get_meds_output_dtype(1)
    for tup in ref_data:
        # Some of these tuples have 3 items, not 2.  The last two are the full dtype tuple.
        name = tup[0]
        if len(tup) == 2:
            dt = tup[1]
        else:
            dt = tup[1:]
        dt = numpy.dtype(dt)
        print(name, dt, cat.dtype[name], dt.char, cat.dtype[name].char)
        assert name in cat.dtype.names, "column %s not present in object_data extension"%name
        assert dt.char == cat.dtype[name].char, "column %s is the wrong type"%name

    # Check that we have the right number of objects.
    n_obj = len(cat)
    print('number of objects is %d' % n_obj)
    numpy.testing.assert_equal(n_obj,n_obj_test,
                               err_msg="MEDS file has wrong number of objects")

    # loop over objects and exposures - test get_cutout
    for iobj in range(n_obj):

        # check ID is correct
        numpy.testing.assert_equal(cat['id'][iobj], iobj+1,
                                   err_msg="MEDS file has wrong id for object %d"%iobj)

        # get number of cutouts and check if it's right
        n_cut = cat['ncutout'][iobj]
        numpy.testing.assert_equal(n_cut,n_cut_test,
                                   err_msg="MEDS file has wrong ncutout for object %d"%iobj)

        # loop over cutouts
        for icut in range(n_cut):

            # get the images etc to compare with originals
            img = m.get_cutout(iobj, icut, type='image')
            wth = m.get_cutout(iobj, icut, type='weight')
            seg = m.get_cutout(iobj, icut, type='seg')
            psf = m.get_psf(iobj, icut)
            wcs_meds = m.get_jacobian(iobj, icut)
            # Note: col == x, row == y.
            wcs_array_meds= numpy.array(
                [ wcs_meds['dudcol'], wcs_meds['dudrow'],
                  wcs_meds['dvdcol'], wcs_meds['dvdrow'],
                  wcs_meds['col0'], wcs_meds['row0'] ] )

            # compare
            numpy.testing.assert_array_equal(img, objlist[iobj].images[icut].array,
                                             err_msg="MEDS cutout has wrong img for object %d"%iobj)
            numpy.testing.assert_array_equal(wth, objlist[iobj].weight[icut].array,
                                             err_msg="MEDS cutout has wrong wth for object %d"%iobj)
            numpy.testing.assert_array_equal(seg, objlist[iobj].seg[icut].array,
                                             err_msg="MEDS cutout has wrong seg for object %d"%iobj)
            numpy.testing.assert_array_equal(psf, objlist[iobj].psf[icut].array,
                                             err_msg="MEDS cutout has wrong psf for object %d"%iobj)
            wcs_orig = objlist[iobj].wcs[icut]
            wcs_array_orig = numpy.array(
                    [ wcs_orig.dudx, wcs_orig.dudy, wcs_orig.dvdx, wcs_orig.dvdy,
                      wcs_orig.origin.x, wcs_orig.origin.y ])
            numpy.testing.assert_array_equal(wcs_array_meds, wcs_array_orig,
                                             err_msg="MEDS cutout has wrong wcs for object %d"%iobj)

        # get the mosaic to compare with originals
        img = m.get_mosaic( iobj, type='image')
        wth = m.get_mosaic( iobj, type='weight')
        seg = m.get_mosaic( iobj, type='seg')
        # There is currently no get_mosaic option for the psfs.
        #psf = m.get_mosaic( iobj, type='psf')
        psf = numpy.concatenate([m.get_psf(iobj,icut) for icut in range(n_cut)], axis=0)

        # get the concatenated images - create the true mosaic
        true_mosaic_img = numpy.concatenate([x.array for x in objlist[iobj].images], axis=0)
        true_mosaic_wth = numpy.concatenate([x.array for x in objlist[iobj].weight], axis=0)
        true_mosaic_seg = numpy.concatenate([x.array for x in objlist[iobj].seg],    axis=0)
        true_mosaic_psf = numpy.concatenate([x.array for x in objlist[iobj].psf],    axis=0)

        # compare
        numpy.testing.assert_array_equal(true_mosaic_img, img,
                                         err_msg="MEDS mosaic has wrong img for object %d"%iobj)
        numpy.testing.assert_array_equal(true_mosaic_wth, wth,
                                         err_msg="MEDS mosaic has wrong wth for object %d"%iobj)
        numpy.testing.assert_array_equal(true_mosaic_seg, seg,
                                         err_msg="MEDS mosaic has wrong seg for object %d"%iobj)
        numpy.testing.assert_array_equal(true_mosaic_psf, psf,
                                         err_msg="MEDS mosaic has wrong psf for object %d"%iobj)



@timer
def test_meds_config(run_slow):
    """
    Create a meds file from a config and compare with a manual creation.
    """
    # Some parameters:
    if run_slow:
        nobj = 5
        n_per_obj = 8
    else:
        nobj = 2
        n_per_obj = 3
    file_name = 'output/test_meds.fits'
    stamp_size = 64
    pixel_scale = 0.26
    seed = 5757231
    g1 = -0.17
    g2 = 0.23

    # generate offsets that depend on the object num so can be easily reproduced
    # for testing below
    offset_x = '$ np.sin(999.*(@obj_num+1))'
    offset_y = '$ np.sin(998.*(@obj_num+1))'
    def get_offset(obj_num):
        return galsim.PositionD(np.sin(999.*(obj_num+1)),np.sin(998.*(obj_num+1)))

    # The config dict to write some images to a MEDS file
    config = {
        'gal' : { 'type' : 'Sersic',
                  'n' : 1.3,
                  'half_light_radius' : { 'type' : 'Sequence', 'first' : 0.7, 'step' : 0.1,
                                          'repeat' : n_per_obj },
                  'shear' : { 'type' : 'G1G2', 'g1' : g1, 'g2' : g2 },
                },
        'psf' : { 'type' : 'Moffat', 'beta' : 2.9, 'fwhm' : 0.7 },
        'image' : { 'pixel_scale' : pixel_scale,
                    'size' : stamp_size, 'random_seed' : seed },
        'output' : { 'type' : 'MEDS',
                     'nobjects' : nobj,
                     'nstamps_per_object' : n_per_obj,
                     'file_name' : file_name
                   }
    }

    import logging
    logging.basicConfig(format="%(message)s", level=logging.WARN, stream=sys.stdout)
    logger = logging.getLogger('test_meds_config')
    galsim.config.BuildFile(galsim.config.CopyConfig(config), logger=logger)

    # Add in badpix and offset so we run both with and without options.
    config = galsim.config.CleanConfig(config)
    config['image']['offset'] = { 'type' : 'XY' , 'x' : offset_x, 'y' : offset_y }
    config['output']['badpix'] = {}
    # These three are just added for coverage really.
    config['output']['weight'] = {}
    config['output']['psf'] = {}
    config['output']['meds_get_offset'] = {}
    galsim.config.BuildFile(galsim.config.CopyConfig(config), logger=logger)

    # Scattered image is invalid with MEDS output
    config = galsim.config.CleanConfig(config)
    config['image'] = {
        'type' : 'Scattered',
        'nobjects' : 20,
        'pixel_scale' : pixel_scale,
        'size' : stamp_size ,
    }
    with assert_raises(galsim.GalSimConfigError):
        galsim.config.BuildFile(galsim.config.CopyConfig(config), logger=logger)

    # Now repeat, making a separate file for each
    config = galsim.config.CleanConfig(config)
    config['gal']['half_light_radius'] = { 'type' : 'Sequence', 'first' : 0.7, 'step' : 0.1,
                                           'index_key' : 'file_num' }
    config['output'] = { 'type' : 'Fits',
                         'nfiles' : nobj,
                         'weight' : { 'hdu' : 1 },
                         'badpix' : { 'hdu' : 2 },
                         'psf' : { 'hdu' : 3 },
                         'dir' : 'output',
                         'file_name' : { 'type' : 'NumberedFile', 'root' : 'test_meds' }
                       }
    config['image'] = { 'type' : 'Tiled',
                        'nx_tiles' : 1,
                        'ny_tiles' : n_per_obj,
                        'pixel_scale' : pixel_scale,
                        'offset' : { 'type' : 'XY' , 'x' : offset_x, 'y' : offset_y },
                        'stamp_size' : stamp_size,
                        'random_seed' : seed
                      }
    galsim.config.Process(galsim.config.CopyConfig(config), logger=logger)

    try:
        import meds
        import fitsio
    except ImportError:
        print('Failed to import either meds or fitsio.  Unable to do tests of meds file.')
        return

    try:
        m = meds.MEDS(file_name)
    except AttributeError:
        print('Seems to be the wrong meds package.  Unable to do tests of meds file.')
        return

    assert m.size == nobj

    # Test that the images made as meds mosaics match the ones written to the separate fits files.
    cat = m.get_cat()
    for iobj in range(nobj):
        ref_file = os.path.join('output','test_meds%d.fits' % iobj)
        ref_im = galsim.fits.read(ref_file)

        meds_im_array = m.get_mosaic(iobj)

        # Just for reference.  If you get an error, you can open this file with ds9.
        alt_meds_file = os.path.join('output','test_alt_meds%d.fits' % iobj)
        alt_meds_im = galsim.Image(meds_im_array)
        alt_meds_im.write(alt_meds_file)

        numpy.testing.assert_array_equal(ref_im.array, meds_im_array,
                                         err_msg="config MEDS has wrong im for object %d"%iobj)

        meds_wt_array = m.get_mosaic(iobj, type='weight')
        ref_wt_im = galsim.fits.read(ref_file, hdu=1)
        numpy.testing.assert_array_equal(ref_wt_im.array, meds_wt_array,
                                         err_msg="config MEDS has wrong wt for object %d"%iobj)

        meds_seg_array = m.get_mosaic(iobj, type='seg')
        ref_seg_im = galsim.fits.read(ref_file, hdu=2)
        ref_seg_im = 1 - ref_seg_im  # The seg mag is 1 where badpix == 0
        numpy.testing.assert_array_equal(ref_seg_im.array, meds_seg_array,
                                         err_msg="config MEDS has wrong seg for object %d"%iobj)

        meds_psf_array = numpy.concatenate([m.get_psf(iobj,icut) for icut in range(n_per_obj)],
                                           axis=0)
        ref_psf_im = galsim.fits.read(ref_file, hdu=3)
        numpy.testing.assert_array_equal(ref_psf_im.array, meds_psf_array,
                                         err_msg="config MEDS has wrong psf for object %d"%iobj)

    # Check that the various positions and sizes are set correctly.
    info = m.get_image_info()
    for iobj in range(nobj):
        n_cut = cat['ncutout'][iobj]
        for icut in range(n_cut):

            # This should be stamp_size
            box_size = cat['box_size'][iobj]
            numpy.testing.assert_almost_equal(box_size, stamp_size)

            # cutout_row and cutout_col are the "zero-offset"
            # position of the object in the stamp. In this convention, the center
            # of the first pixel is at (0,0), call this meds_center.
            # This means cutout_row/col should be the same as meds_center + offset
            offset = get_offset(iobj*n_cut+icut)
            meds_center = galsim.PositionD( (box_size-1.)/2., (box_size-1.)/2. )
            cutout_row = cat['cutout_row'][iobj][icut]
            cutout_col = cat['cutout_col'][iobj][icut]
            print('cutout_row, cutout_col = ',cutout_col, cutout_row)
            numpy.testing.assert_almost_equal(cutout_col,
                                              (meds_center+offset).x)
            numpy.testing.assert_almost_equal(cutout_row,
                                              (meds_center+offset).y)

            # The col0 and row0 here should be the same.
            wcs_meds = m.get_jacobian(iobj, icut)
            numpy.testing.assert_almost_equal(wcs_meds['col0'],
                                              (meds_center+offset).x)
            numpy.testing.assert_almost_equal(wcs_meds['row0'],
                                              (meds_center+offset).y)


            # The centroid should be (roughly) at the nominal center + offset
            img = m.get_cutout(iobj, icut, type='image')
            x,y = numpy.meshgrid( range(img.shape[1]), range(img.shape[0]) )
            itot = numpy.sum(img)
            ix = numpy.sum(x*img)
            iy = numpy.sum(y*img)
            print('centroid = ',ix/itot, iy/itot)

            print('center + offset = ',meds_center + offset)
            numpy.testing.assert_almost_equal(ix/itot,
                                              (meds_center+offset).x, decimal=2)
            numpy.testing.assert_almost_equal(iy/itot,
                                              (meds_center+offset).y, decimal=2)

            # The orig positions are irrelevant and should be 0.
            orig_row = cat['orig_row'][iobj][icut]
            orig_col = cat['orig_col'][iobj][icut]
            orig_start_row = cat['orig_start_row'][iobj][icut]
            orig_start_col = cat['orig_start_col'][iobj][icut]
            numpy.testing.assert_almost_equal(orig_col, 0.)
            numpy.testing.assert_almost_equal(orig_row, 0.)
            numpy.testing.assert_almost_equal(orig_start_col, 0.)
            numpy.testing.assert_almost_equal(orig_start_row, 0.)

            # This should be also be 0.
            numpy.testing.assert_almost_equal(info['position_offset'], 0.)


@timer
def test_nan_fits():
    """Test reading in a FITS file that has NAN.0 entries in the header.

    This test is specifically in response to issue #602.
    """
    import warnings
    from galsim._pyfits import pyfits
    # Older pyfits versions don't have this, so just skip this test then.
    if not hasattr(pyfits, 'verify'): return

    # The problematic file:
    file_name = "des_data/DECam_00158414_01.fits.fz"

    # These are the values we should be reading in:
    ref_bounds = galsim.BoundsI(xmin=1, xmax=2048, ymin=1, ymax=4096)
    ref_wcs = galsim.GSFitsWCS(_data = [
            'TPV',
            numpy.array([13423.2, 6307.333]),
            numpy.array([[-4.410051713005e-09, 7.286844513153e-05],
                   [-7.285161461796e-05, 3.936353853081e-09]]),
            galsim.CelestialCoord(1.1502513773465992 * galsim.radians,
                                  -0.9862866578241959 * galsim.radians),
            numpy.array(
                    [[[0.004336243600183, -0.01133740904139, 0.01202041999278, -0.004357212119479],
                      [1.013741474567, -0.01657049389296, 0.005805882078771, 0.0],
                      [0.008865811106037, -0.007472254968395, 0.0, 0.0],
                      [0.0008534196190617, 0.0, 0.0, 0.0]],
                     [[0.002619866608142, 0.9931356822158, 0.008771460618847, -0.003739430249945],
                      [-0.009422336649176, 0.01826140592329, -0.009387805146152, 0.0],
                      [-0.01066967054507, 0.007202907073747, 0.0, 0.0],
                      [-0.003683686751425, 0.0, 0.0, 0.0]]
                    ]),
            None, None])

    # First just read the file directly, not using galsim.fits.read
    with pyfits.open(file_name) as fp:
        try:
            data = fp[1].data
            print('Able to read FITS file with NAN.0 without any problem.')
        except:
            print('Running verify to fix the problematic FITS header.')
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore",category=pyfits.verify.VerifyWarning)
                fp[1].verify('fix')
            # This should work now.
            data = fp[1].data
        header = fp[1].header

    assert data.shape == ref_bounds.numpyShape()

    # Check a direct read of the header with GSFitsWCS
    wcs = galsim.GSFitsWCS(header=header)
    assert wcs == ref_wcs

    # Now read it with GalSim's fits.read function.
    # Reading this file will emit verification warnings, so we'll ignore those here for the
    # test.  But the result should be a valid image.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",category=pyfits.verify.VerifyWarning)
        im = galsim.fits.read(file_name)

    assert im.bounds == ref_bounds
    assert im.wcs == ref_wcs


@timer
def test_psf():
    """Test the two kinds of PSF files we have in DES.
    """
    data_dir = 'des_data'
    psfex_file = "DECam_00154912_12_psfcat.psf"
    fitpsf_file = "DECam_00154912_12_fitpsf.fits"
    wcs_file = "DECam_00154912_12_header.fits"

    wcs = galsim.FitsWCS(wcs_file, dir=data_dir)

    # We don't require that the files in example_data_dir have been downloaded.  If they
    # haven't, then we just directly set the comparison values that we want here.
    example_data_dir = '../examples/des/des_data'
    cat_file = "DECam_00154912_12_cat.fits"
    image_file = "DECam_00154912_12.fits.fz"

    try:
        cat = galsim.Catalog(cat_file, hdu=2, dir=example_data_dir)
        size = numpy.array([ cat.getFloat(i,'FLUX_RADIUS') for i in range(cat.nobjects) ])
        mag = numpy.array([ cat.getFloat(i,'MAG_AUTO') for i in range(cat.nobjects) ])
        flags = numpy.array([ cat.getInt(i,'FLAGS') for i in range(cat.nobjects) ])
        index = numpy.array(range(cat.nobjects))
        xvals = numpy.array([ cat.getFloat(i,'X_IMAGE') for i in range(cat.nobjects) ])
        yvals = numpy.array([ cat.getFloat(i,'Y_IMAGE') for i in range(cat.nobjects) ])

        # Pick bright small objects as probable stars
        mask = (flags == 0) & (mag < 14) & (mag > 13) & (size > 2) & (size < 2.5)
        idx = numpy.argsort(size[mask])

        # This choice of a star is fairly isolated from neighbors, isn't too near an edge or a tape
        # bump, and doesn't have any noticeable image artifacts in its vicinity.
        x = xvals[mask][idx][27]
        y = yvals[mask][idx][27]
        print('Using x,y = ',x,y)
        image_pos = galsim.PositionD(x,y)
        print('size, mag = ',size[mask][idx][27], mag[mask][idx][27])

        data = galsim.fits.read(image_file, dir=example_data_dir)
        b = galsim.BoundsI(int(x)-15, int(x)+16, int(y)-15, int(y)+16)
        data_stamp = data[b]

        header = galsim.fits.FitsHeader(image_file, dir=example_data_dir)
        sky_level = header['SKYBRITE']
        data_stamp -= sky_level

        raw_meas = data_stamp.FindAdaptiveMom()
        print('raw_meas = ',raw_meas)
        ref_size = raw_meas.moments_sigma
        ref_shape = raw_meas.observed_shape
        print('ref size: ',ref_size)
        print('ref shape: ',ref_shape)

    except OSError:
        x,y = 1195.64074707, 1276.63427734
        image_pos = galsim.PositionD(x,y)
        b = galsim.BoundsI(int(x)-15, int(x)+16, int(y)-15, int(y)+16)
        ref_size = 1.80668628216
        ref_shape = galsim.Shear(g1=0.022104322221,g2=-0.130925191715)

    # First the PSFEx model using the wcs_file to get the model is sky coordinates.
    psfex = galsim.des.DES_PSFEx(psfex_file, wcs_file, dir=data_dir)
    psf = psfex.getPSF(image_pos)

    # The getLocalWCS function should return a local WCS
    assert psfex.getLocalWCS(image_pos).isLocal()

    # Draw the postage stamp image
    # Note: the PSF already includes the pixel response, so draw with method 'no_pixel'.
    stamp = psf.drawImage(wcs=wcs.local(image_pos), bounds=b, method='no_pixel')
    print('wcs = ',wcs.local(image_pos))
    meas = stamp.FindAdaptiveMom()
    print('meas = ',meas)
    print('pixel scale = ',stamp.wcs.minLinearScale(image_pos=image_pos))
    print('cf sizes: ',ref_size, meas.moments_sigma)
    print('cf shapes: ',ref_shape, meas.observed_shape)
    # The agreement for a single star is not great of course, not even 2 decimals.
    # Divide by 2 to get agreement at 2 dp.
    numpy.testing.assert_almost_equal(meas.moments_sigma/2, ref_size/2, decimal=2,
                                      err_msg="PSFEx size doesn't match")
    numpy.testing.assert_almost_equal(meas.observed_shape.g1/2, ref_shape.g1/2, decimal=2,
                                      err_msg="PSFEx shape.g1 doesn't match")
    numpy.testing.assert_almost_equal(meas.observed_shape.g2/2, ref_shape.g2/2, decimal=2,
                                      err_msg="PSFEx shape.g2 doesn't match")

    # Repeat without the wcs_file argument, so the model is in chip coordinates.
    # Also check the functionality where the file is already open.
    with pyfits.open(os.path.join(data_dir, psfex_file)) as hdu_list:
        psfex = galsim.des.DES_PSFEx(hdu_list[1])
    psf = psfex.getPSF(image_pos)

    # In this case, the getLocalWCS function won't return anything useful.
    assert psfex.getLocalWCS(image_pos) is None

    # Draw the postage stamp image.  This time in image coords, so pixel_scale = 1.0.
    stamp = psf.drawImage(bounds=b, scale=1.0, method='no_pixel')
    meas = stamp.FindAdaptiveMom()
    numpy.testing.assert_almost_equal(meas.moments_sigma/2, ref_size/2, decimal=2,
                                      err_msg="no-wcs PSFEx size doesn't match")
    numpy.testing.assert_almost_equal(meas.observed_shape.g1/2, ref_shape.g1/2, decimal=2,
                                      err_msg="no-wcs PSFEx shape.g1 doesn't match")
    numpy.testing.assert_almost_equal(meas.observed_shape.g2/2, ref_shape.g2/2, decimal=2,
                                      err_msg="no-wcs PSFEx shape.g2 doesn't match")

    with assert_raises(TypeError):
        # file_name must be a string.
        galsim.des.DES_PSFEx(psf, wcs=wcs_file, dir=data_dir)
    with assert_raises(galsim.GalSimError):
        # Cannot provide both image_file_name and wcs
        galsim.des.DES_PSFEx(psfex_file, image_file_name=wcs_file, wcs=wcs_file, dir=data_dir)
    with assert_raises(OSError):
        # This one doesn't exist.
        galsim.des.DES_PSFEx('nonexistant.psf', wcs=wcs_file, dir=data_dir)
    with assert_raises(OSError):
        # This one exists, but has invalid header parameters.
        galsim.des.DES_PSFEx('invalid_psfcat.psf', wcs=wcs_file, dir=data_dir)

    # Now the shapelet PSF model.  This model is already in sky coordinates, so no wcs_file needed.
    fitpsf = galsim.des.DES_Shapelet(os.path.join(data_dir,fitpsf_file))
    psf = fitpsf.getPSF(image_pos)

    # Draw the postage stamp image
    # Again, the PSF already includes the pixel response.
    stamp = psf.drawImage(wcs=wcs.local(image_pos), bounds=b, method='no_pixel')
    meas = stamp.FindAdaptiveMom()
    numpy.testing.assert_almost_equal(meas.moments_sigma/2, ref_size/2, decimal=2,
                                      err_msg="Shapelet PSF size doesn't match")
    numpy.testing.assert_almost_equal(meas.observed_shape.g1/2, ref_shape.g1/2, decimal=2,
                                      err_msg="Shapelet PSF shape.g1 doesn't match")
    numpy.testing.assert_almost_equal(meas.observed_shape.g2/2, ref_shape.g2/2, decimal=2,
                                      err_msg="Shapelet PSF shape.g2 doesn't match")

    with assert_raises(galsim.GalSimBoundsError):
        fitpsf.getPSF(image_pos = galsim.PositionD(4000, 5000))


@timer
def test_psf_config():
    """Test building the two PSF types using the config layer.
    """
    data_dir = 'des_data'
    psfex_file = "DECam_00154912_12_psfcat.psf"
    fitpsf_file = "DECam_00154912_12_fitpsf.fits"
    wcs_file = "DECam_00154912_12_header.fits"

    image_pos = galsim.PositionD(123.45, 543.21)

    config = {
        'input' : {
            'des_shapelet' : { 'dir' : data_dir, 'file_name' : fitpsf_file },
            'des_psfex' : [
                { 'dir' : data_dir, 'file_name' : psfex_file },
                { 'dir' : data_dir, 'file_name' : psfex_file, 'image_file_name' : wcs_file },
            ]
        },

        'psf1' : { 'type' : 'DES_Shapelet' },
        'psf2' : { 'type' : 'DES_PSFEx', 'num' : 0 },
        'psf3' : { 'type' : 'DES_PSFEx', 'num' : 1 },
        'psf4' : { 'type' : 'DES_Shapelet', 'image_pos' : galsim.PositionD(567,789), 'flux' : 179,
                   'gsparams' : { 'folding_threshold' : 1.e-4 } },
        'psf5' : { 'type' : 'DES_PSFEx', 'image_pos' : galsim.PositionD(789,567), 'flux' : 388,
                   'gsparams' : { 'folding_threshold' : 1.e-4 } },
        'bad1' : { 'type' : 'DES_Shapelet', 'image_pos' : galsim.PositionD(5670,789) },

        # This would normally be set by the config processing.  Set it manually here.
        'image_pos' : image_pos,
    }

    galsim.config.ProcessInput(config)

    psf1a = galsim.config.BuildGSObject(config, 'psf1')[0]
    fitpsf = galsim.des.DES_Shapelet(fitpsf_file, dir=data_dir)
    psf1b = fitpsf.getPSF(image_pos)
    gsobject_compare(psf1a, psf1b)

    psf2a = galsim.config.BuildGSObject(config, 'psf2')[0]
    psfex0 = galsim.des.DES_PSFEx(psfex_file, dir=data_dir)
    psf2b = psfex0.getPSF(image_pos)
    gsobject_compare(psf2a, psf2b)

    psf3a = galsim.config.BuildGSObject(config, 'psf3')[0]
    psfex1 = galsim.des.DES_PSFEx(psfex_file, wcs_file, dir=data_dir)
    psf3b = psfex1.getPSF(image_pos)
    gsobject_compare(psf3a, psf3b)

    gsparams = galsim.GSParams(folding_threshold=1.e-4)
    psf4a = galsim.config.BuildGSObject(config, 'psf4')[0]
    psf4b = fitpsf.getPSF(galsim.PositionD(567,789),gsparams=gsparams).withFlux(179)
    gsobject_compare(psf4a, psf4b)

    # Insert a wcs for thes last one.
    config['wcs'] = galsim.FitsWCS(os.path.join(data_dir,wcs_file))
    config = galsim.config.CleanConfig(config)
    galsim.config.ProcessInput(config)
    psfex2 = galsim.des.DES_PSFEx(psfex_file, dir=data_dir, wcs=config['wcs'])
    psf5a = galsim.config.BuildGSObject(config, 'psf5')[0]
    psf5b = psfex2.getPSF(galsim.PositionD(789,567),gsparams=gsparams).withFlux(388)
    gsobject_compare(psf5a, psf5b)

    del config['image_pos']
    galsim.config.RemoveCurrent(config)
    with assert_raises(galsim.GalSimConfigError):
        galsim.config.BuildGSObject(config, 'psf1')[0]
    with assert_raises(galsim.GalSimConfigError):
        galsim.config.BuildGSObject(config, 'psf2')[0]
    with assert_raises(galsim.config.gsobject.SkipThisObject):
        galsim.config.BuildGSObject(config, 'bad1')[0]


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

"""Unit tests for the PSF correction and shear estimation routines.

There are two types of tests: tests that use Gaussian profiles, for which the ideal results are
known; and tests that use real galaxies in SDSS for which results were tabulated using the same code
before it was integrated into GalSim (so we can make sure we are not breaking anything as we modify
the code).
"""

import os
import numpy as np
import math
import glob

import galsim
from galsim_test_helpers import *


# define a range of input parameters for the Gaussians that we are testing
gaussian_sig_values = [0.5, 1.0, 2.0]
shear_values = [0.01, 0.03, 0.10, 0.30]
pixel_scale = 0.2
decimal = 2 # decimal place at which to require equality in sizes
decimal_shape = 3 # decimal place at which to require equality in shapes

# The timing tests can be unreliable in environments with other processes running at the
# same time.  So we disable them by default.  However, on a clean system, they should all pass.
test_timing = False

# define inputs and expected results for tests that use real SDSS galaxies
img_dir = os.path.join(".","HSM_precomputed")
gal_file_prefix = "image."
psf_file_prefix = "psf."
img_suff = ".fits"
file_indices = [0, 2, 4, 6, 8]
x_centroid = [35.888, 19.44, 8.74, 20.193, 57.94]
y_centroid = [19.845, 25.047, 11.92, 38.93, 27.73]
sky_var = [35.01188, 35.93418, 35.15456, 35.11146, 35.16454]
correction_methods = ["KSB", "BJ", "LINEAR", "REGAUSS"]
# Note: expected results give shear for KSB and distortion for others, but the results below have
# converted KSB expected results to distortion for the sake of consistency
e1_expected = np.array([
        [0.467603106752, 0.381211727, 0.398856937, 0.401755571],
        [0.28618443944, 0.199222784, 0.233883543, 0.234257525],
        [0.271533794146, 0.158049396, 0.183517068, 0.184893412],
        [-0.293754156071, -0.457024541, 0.123946584, -0.609233462],
        [0.557720893779, 0.374143023, 0.714147448, 0.435404409] ])
e2_expected = np.array([
        [-0.867225166489, -0.734855778, -0.777027588, -0.774684891],
        [-0.469354341577, -0.395520479, -0.502540961, -0.464466257],
        [-0.519775291311, -0.471589061, -0.574750641, -0.529664935],
        [0.345688365839, -0.342047099, 0.120603755, -0.44609129428863525],
        [0.525728304099, 0.370691830, 0.702724807, 0.433999442] ])
resolution_expected = np.array([
        [0.796144249, 0.835624917, 0.835624917, 0.827796187],
        [0.685023735, 0.699602704, 0.699602704, 0.659457638],
        [0.634736458, 0.651040481, 0.651040481, 0.614663396],
        [0.477027015, 0.477210752, 0.477210752, 0.423157447],
        [0.595205998, 0.611824797, 0.611824797, 0.563582092] ])
sigma_e_expected = np.array([
        [0.016924826, 0.014637648, 0.014637648, 0.014465546],
        [0.075769504, 0.073602324, 0.073602324, 0.064414520],
        [0.110253112, 0.106222900, 0.106222900, 0.099357106],
        [0.185276702, 0.184300955, 0.184300955, 0.173478300],
        [0.073020065, 0.070270966, 0.070270966, 0.061856263] ])

def equal_hsmshapedata(res1, res2):
    """Utility to check that all entries in two ShapeData objects are equal."""
    if not isinstance(res1, galsim.hsm.ShapeData) or not isinstance(res2, galsim.hsm.ShapeData):
        raise TypeError("Objects that were passed in are not ShapeData objects")

    if res1.corrected_e1 != res2.corrected_e1: return False
    if res1.corrected_e2 != res2.corrected_e2: return False
    if res1.corrected_g1 != res2.corrected_g1: return False
    if res1.corrected_g2 != res2.corrected_g2: return False
    if res1.corrected_shape_err != res2.corrected_shape_err: return False
    if res1.error_message != res2.error_message: return False
    if res1.image_bounds.xmin != res2.image_bounds.xmin: return False
    if res1.image_bounds.xmax != res2.image_bounds.xmax: return False
    if res1.image_bounds.ymin != res2.image_bounds.ymin: return False
    if res1.image_bounds.ymax != res2.image_bounds.ymax: return False
    if res1.meas_type != res2.meas_type: return False
    if res1.moments_amp != res2.moments_amp: return False
    if res1.moments_centroid != res2.moments_centroid: return False
    if res1.moments_n_iter != res2.moments_n_iter: return False
    if res1.moments_rho4 != res2.moments_rho4: return False
    if res1.moments_sigma != res2.moments_sigma: return False
    if res1.moments_status != res2.moments_status: return False
    if res1.observed_shape != res2.observed_shape: return False
    if res1.resolution_factor != res2.resolution_factor: return False
    return True


@timer
def test_moments_basic():
    """Test that we can properly recover adaptive moments for Gaussians."""
    first_test=True
    for sig in gaussian_sig_values:
        for g1 in shear_values:
            for g2 in shear_values:
                total_shear = np.sqrt(g1**2 + g2**2)
                conversion_factor = np.tanh(2.0*math.atanh(total_shear))/total_shear
                distortion_1 = g1*conversion_factor
                distortion_2 = g2*conversion_factor
                gal = galsim.Gaussian(flux = 1.0, sigma = sig)
                gal = gal.shear(g1=g1, g2=g2)
                gal_image = gal.drawImage(scale = pixel_scale, method='no_pixel')
                result = gal_image.FindAdaptiveMom()
                # make sure we find the right Gaussian sigma
                np.testing.assert_almost_equal(np.fabs(result.moments_sigma-sig/pixel_scale), 0.0,
                                               err_msg = "- incorrect dsigma", decimal = decimal)
                # make sure we find the right e
                np.testing.assert_almost_equal(result.observed_shape.e1,
                                               distortion_1, err_msg = "- incorrect e1",
                                               decimal = decimal_shape)
                np.testing.assert_almost_equal(result.observed_shape.e2,
                                               distortion_2, err_msg = "- incorrect e2",
                                               decimal = decimal_shape)

                # test for moments with a circular weight function
                result_round = gal_image.FindAdaptiveMom(round_moments=True)

                # The ellipticities calculated here are we expect when integrating an
                # elliptical Gaussian light profile with a round Gaussian weight function
                # with the same sigma.
                q = np.exp(-2.*math.atanh(total_shear))
                theta = 0.5*math.atan2(g2, g1)
                e_round = (1 - q**2)/(1 + q**2 + 2*q)
                e1_round = e_round*math.cos(2*theta)
                e2_round = e_round*math.sin(2*theta)

                np.testing.assert_almost_equal(np.fabs(result_round.moments_sigma-sig/pixel_scale), 0.0,
                                               err_msg = "- incorrect round sigma", decimal = decimal)
                np.testing.assert_almost_equal(result_round.observed_shape.e1,
                                               e1_round, err_msg = "- incorrect round e1",
                                               decimal = decimal_shape)
                np.testing.assert_almost_equal(result_round.observed_shape.e2,
                                               e2_round, err_msg = "- incorrect round e2",
                                               decimal = decimal_shape)

                # if this is the first time through this loop, just make sure it runs and gives the
                # same result whether const or not.
                if first_test:
                    result = gal_image.view().FindAdaptiveMom()
                    first_test=False
                    np.testing.assert_almost_equal(
                        np.fabs(result.moments_sigma-sig/pixel_scale), 0.0,
                        err_msg = "- incorrect dsigma (ImageView)", decimal = decimal)
                    np.testing.assert_almost_equal(
                        result.observed_shape.e1,
                        distortion_1, err_msg = "- incorrect e1 (ImageView)",
                        decimal = decimal_shape)
                    np.testing.assert_almost_equal(
                        result.observed_shape.e2,
                        distortion_2, err_msg = "- incorrect e2 (ImageView)",
                        decimal = decimal_shape)
                    result = gal_image.view(make_const=True).FindAdaptiveMom()
                    np.testing.assert_almost_equal(
                        np.fabs(result.moments_sigma-sig/pixel_scale), 0.0,
                        err_msg = "- incorrect dsigma (make_const=True)", decimal = decimal)
                    np.testing.assert_almost_equal(
                        result.observed_shape.e1,
                        distortion_1, err_msg = "- incorrect e1 (make_const=True)",
                        decimal = decimal_shape)
                    np.testing.assert_almost_equal(
                        result.observed_shape.e2,
                        distortion_2, err_msg = "- incorrect e2 (make_const=True)",
                        decimal = decimal_shape)

@timer
def test_moments_wcs():
    """Test adaptive moments for Gaussians when the wcs is non-trivial."""

    wcs_list = [
        galsim.PixelScale(0.2),
        galsim.JacobianWCS(0.2, 0.03, -0.03, 0.2),
        galsim.JacobianWCS(0.23, 0.03, -0.06, 0.19),
        galsim.AffineTransform(0.23, 0.03, 0.06, -0.19,
                               origin=galsim.PositionD(75,105),
                               world_origin=galsim.PositionD(3475,1005)),
        galsim.TanWCS(galsim.AffineTransform(0.03, 0.23, 0.21, -0.01),
                      galsim.CelestialCoord(74.2 * galsim.degrees, -32 * galsim.degrees)),
    ]

    rng = galsim.BaseDeviate(1234)

    for sig in gaussian_sig_values:
        for g1 in shear_values:
            for g2 in shear_values:
                for wcs in wcs_list:
                    du = rng.np.uniform(-2,2)
                    dv = rng.np.uniform(-2,2)
                    gal = galsim.Gaussian(sigma=sig).shear(g1=g1, g2=g2).shift(du,dv)
                    image = gal.drawImage(nx=200, ny=200, wcs=wcs, method='no_pixel')
                    result = image.FindAdaptiveMom()

                    # Convert to sky coords
                    result = result.applyWCS(wcs, image_pos=image.true_center)
                    print(result.moments_sigma, sig,
                          result.observed_shape.g1, g1,
                          result.observed_shape.g2, g2,
                          result.moments_centroid.x, du,
                          result.moments_centroid.y, dv)
                    np.testing.assert_allclose(result.moments_sigma, sig, rtol=1.e-5)
                    np.testing.assert_allclose(result.observed_shape.g1, g1, rtol=1.e-5)
                    np.testing.assert_allclose(result.observed_shape.g2, g2, rtol=1.e-5)
                    np.testing.assert_allclose(result.moments_centroid.x, du, atol=1.e-7)
                    np.testing.assert_allclose(result.moments_centroid.y, dv, atol=1.e-7)

                    # Can also do this directly with FindAdaptiveMom(use_sky_coords=True)
                    result = image.FindAdaptiveMom(use_sky_coords=True)
                    np.testing.assert_allclose(result.moments_sigma, sig, rtol=1.e-5)
                    np.testing.assert_allclose(result.observed_shape.g1, g1, rtol=1.e-5)
                    np.testing.assert_allclose(result.observed_shape.g2, g2, rtol=1.e-5)
                    np.testing.assert_allclose(result.moments_centroid.x, du, atol=1.e-7)
                    np.testing.assert_allclose(result.moments_centroid.y, dv, atol=1.e-7)


@timer
def test_shearest_basic():
    """Test that we can recover shears for Gaussian galaxies and PSFs."""
    for sig in gaussian_sig_values:
        for g1 in shear_values:
            for g2 in shear_values:
                total_shear = np.sqrt(g1**2 + g2**2)
                conversion_factor = np.tanh(2.0*math.atanh(total_shear))/total_shear
                distortion_1 = g1*conversion_factor
                distortion_2 = g2*conversion_factor
                gal = galsim.Gaussian(flux = 1.0, sigma = sig)
                psf = galsim.Gaussian(flux = 1.0, sigma = sig)
                gal = gal.shear(g1=g1, g2=g2)
                psf = psf.shear(g1=0.1*g1, g2=0.05*g2)
                final = galsim.Convolve([gal, psf])
                final_image = final.drawImage(scale=pixel_scale, method='no_pixel')
                epsf_image = psf.drawImage(scale=pixel_scale, method='no_pixel')
                result = galsim.hsm.EstimateShear(final_image, epsf_image)
                # make sure we find the right e after PSF correction
                # with regauss, which returns a distortion
                np.testing.assert_almost_equal(result.corrected_e1,
                                               distortion_1, err_msg = "- incorrect e1",
                                               decimal = decimal_shape)
                np.testing.assert_almost_equal(result.corrected_e2,
                                               distortion_2, err_msg = "- incorrect e2",
                                               decimal = decimal_shape)


@timer
def test_shearest_precomputed():
    """Test that we can recover shears the same as before the code was put into GalSim."""
    # loop over real galaxies
    for index in range(len(file_indices)):
        # define input filenames
        img_file = os.path.join(img_dir, gal_file_prefix + str(file_indices[index]) + img_suff)
        psf_file = os.path.join(img_dir, psf_file_prefix + str(file_indices[index]) + img_suff)

        # read in information for objects and expected results
        imgR = galsim.fits.read(img_file)
        # perform a cast to int as the images on file are unsigned,
        # which leads to problems when we subtract 1000 below
        img  = galsim.Image(imgR, dtype = int)
        img -= 1000
        psfR = galsim.fits.read(psf_file)
        psf = galsim.Image(psfR, dtype = np.int16)
        psf -= 1000

        # get PSF moments for later tests
        psf_mom = psf.FindAdaptiveMom()

        # loop over methods
        for method_index in range(len(correction_methods)):
            # call PSF correction
            result = galsim.hsm.EstimateShear(
                img, psf, sky_var = sky_var[index], shear_est = correction_methods[method_index],
                guess_centroid = galsim.PositionD(x_centroid[index], y_centroid[index]))

            # compare results with precomputed
            print(result.meas_type, correction_methods[method_index])
            if result.meas_type == 'e':
                np.testing.assert_almost_equal(
                    result.corrected_e1, e1_expected[index][method_index], decimal = decimal_shape)
                np.testing.assert_almost_equal(
                    result.corrected_e2, e2_expected[index][method_index], decimal = decimal_shape)
            else:
                gval = np.sqrt(result.corrected_g1**2 + result.corrected_g2**2)
                if gval <= 1.0:
                    s = galsim.Shear(g1=result.corrected_g1, g2=result.corrected_g2)
                    np.testing.assert_almost_equal(
                        s.e1, e1_expected[index][method_index], decimal = decimal_shape)
                    np.testing.assert_almost_equal(
                        s.e2, e2_expected[index][method_index], decimal = decimal_shape)
            # also compare resolutions and estimated errors
            np.testing.assert_almost_equal(
                result.resolution_factor, resolution_expected[index][method_index],
                decimal = decimal_shape)
            np.testing.assert_almost_equal(
                result.corrected_shape_err, sigma_e_expected[index][method_index],
                decimal = decimal_shape)
            # Also check that the PSF properties that come out of EstimateShear are the same as
            # what we would get from measuring directly.
            np.testing.assert_almost_equal(
                psf_mom.moments_sigma, result.psf_sigma, decimal=decimal_shape,
                err_msg = "PSF sizes from FindAdaptiveMom vs. EstimateShear disagree")
            np.testing.assert_almost_equal(
                psf_mom.observed_shape.e1, result.psf_shape.e1, decimal=decimal_shape,
                err_msg = "PSF e1 from FindAdaptiveMom vs. EstimateShear disagree")
            np.testing.assert_almost_equal(
                psf_mom.observed_shape.e2, result.psf_shape.e2, decimal=decimal_shape,
                err_msg = "PSF e2 from FindAdaptiveMom vs. EstimateShear disagree")
            first = False


@timer
def test_masks():
    """Test that moments and shear estimation routines respond appropriately to masks."""
    # set up some toy galaxy and PSF
    my_sigma = 1.0
    my_pixscale = 0.1
    my_g1 = 0.15
    my_g2 = -0.4
    imsize = 256
    g = galsim.Gaussian(sigma = my_sigma)
    p = galsim.Gaussian(sigma = my_sigma) # the ePSF is Gaussian (kind of silly but it means we can
                                     # predict results exactly)
    g = g.shear(g1=my_g1, g2=my_g2)
    obj = galsim.Convolve(g, p)
    im = galsim.ImageF(imsize, imsize)
    p_im = galsim.ImageF(imsize, imsize)
    obj.drawImage(image=im, scale=my_pixscale, method='no_pixel')
    p.drawImage(image=p_im, scale=my_pixscale, method='no_pixel')

    # make some screwy weight and badpix images that should cause issues, and check that the
    # exception is thrown
    good_weight_im = galsim.ImageI(imsize, imsize, init_value=1)
    ## different size from image
    weight_im = galsim.ImageI(imsize, 2*imsize)
    assert_raises(ValueError, galsim.hsm.FindAdaptiveMom, im, weight_im)
    assert_raises(ValueError, galsim.hsm.EstimateShear, im, p_im, weight_im)
    badpix_im = galsim.ImageI(imsize, 2*imsize)
    assert_raises(ValueError, galsim.hsm.FindAdaptiveMom, im, badpix_im)
    assert_raises(ValueError, galsim.hsm.EstimateShear, im, p_im, good_weight_im, badpix_im)
    ## weird values
    weight_im = galsim.ImageI(imsize, imsize, init_value = -3)
    assert_raises(ValueError, galsim.hsm.FindAdaptiveMom, im, weight_im)
    assert_raises(ValueError, galsim.hsm.EstimateShear, im, p_im, weight_im)
    ## excludes all pixels
    weight_im = galsim.ImageI(imsize, imsize)
    assert_raises(galsim.GalSimError, galsim.hsm.FindAdaptiveMom, im, weight_im)
    assert_raises(galsim.GalSimError, galsim.hsm.EstimateShear, im, p_im, weight_im)
    badpix_im = galsim.ImageI(imsize, imsize, init_value = -1)
    assert_raises(galsim.GalSimError, galsim.hsm.FindAdaptiveMom, im, good_weight_im, badpix_im)
    assert_raises(galsim.GalSimError, galsim.hsm.EstimateShear, im, p_im, good_weight_im, badpix_im)

    # check moments, shear without mask
    resm = im.FindAdaptiveMom()
    ress = galsim.hsm.EstimateShear(im, p_im)

    # check moments, shear with weight image that includes all pixels
    weightall1 = galsim.ImageI(imsize, imsize, init_value = 1)
    resm_weightall1 = im.FindAdaptiveMom(weightall1)
    ress_weightall1 = galsim.hsm.EstimateShear(im, p_im, weightall1)

    # We'll do this series of tests a few times, so encapsulate the code here.
    def check_equal(resm, ress, resm_test, ress_test, tag):
        np.testing.assert_equal(resm.observed_shape.e1, resm_test.observed_shape.e1,
            err_msg="e1 from FindAdaptiveMom changes "+tag)
        np.testing.assert_equal(resm.observed_shape.e2, resm_test.observed_shape.e2,
            err_msg="e2 from FindAdaptiveMom changes "+tag)
        np.testing.assert_equal(resm.moments_sigma, resm_test.moments_sigma,
            err_msg="sigma from FindAdaptiveMom changes "+tag)
        np.testing.assert_equal(ress.observed_shape.e1, ress_test.observed_shape.e1,
            err_msg="observed e1 from EstimateShear changes "+tag)
        np.testing.assert_equal(ress.observed_shape.e2, ress_test.observed_shape.e2,
            err_msg="observed e2 from EstimateShear changes "+tag)
        np.testing.assert_equal(ress.moments_sigma, ress_test.moments_sigma,
            err_msg="observed sigma from EstimateShear changes "+tag)
        np.testing.assert_equal(ress.corrected_e1, ress_test.corrected_e1,
            err_msg="corrected e1 from EstimateShear changes "+tag)
        np.testing.assert_equal(ress.corrected_e2, ress_test.corrected_e2,
            err_msg="corrected e2 from EstimateShear changes "+tag)
        np.testing.assert_equal(ress.resolution_factor, ress_test.resolution_factor,
            err_msg="resolution factor from EstimateShear changes "+tag)
    check_equal(resm,ress,resm_weightall1,ress_weightall1, "when using inclusive weight")

    # check moments and shears with mask of edges, should be nearly the same
    # (this seems dumb, but it's helpful for keeping track of whether the pointers in the C++ code
    # are being properly updated despite the masks.  If we monkey in that code again, it will be a
    # useful check.)
    maskedge = galsim.ImageI(imsize, imsize, init_value = 1)
    xmin = maskedge.xmin
    xmax = maskedge.xmax
    ymin = maskedge.ymin
    ymax = maskedge.ymax
    edgenum = 3
    for ind1 in range(xmin, xmax+1):
        for ind2 in range(ymin, ymax+1):
            if (ind1 <= (xmin+edgenum)) or (ind1 >= (xmax-edgenum)) or (ind2 <= (ymin+edgenum)) or (ind2 >= (ymax-edgenum)):
                maskedge.setValue(ind1, ind2, 0)
    resm_maskedge = im.FindAdaptiveMom(maskedge)
    ress_maskedge = galsim.hsm.EstimateShear(im, p_im, maskedge)
    test_decimal = 4
    np.testing.assert_almost_equal(resm.observed_shape.e1, resm_maskedge.observed_shape.e1,
        decimal=test_decimal, err_msg="e1 from FindAdaptiveMom changes when masking edge")
    np.testing.assert_almost_equal(resm.observed_shape.e2, resm_maskedge.observed_shape.e2,
        decimal=test_decimal, err_msg="e2 from FindAdaptiveMom changes when masking edge")
    np.testing.assert_almost_equal(resm.moments_sigma, resm_maskedge.moments_sigma,
        decimal=test_decimal, err_msg="sigma from FindAdaptiveMom changes when masking edge")
    np.testing.assert_almost_equal(ress.observed_shape.e1, ress_maskedge.observed_shape.e1,
        decimal=test_decimal, err_msg="observed e1 from EstimateShear changes when masking edge")
    np.testing.assert_almost_equal(ress.observed_shape.e2, ress_maskedge.observed_shape.e2,
        decimal=test_decimal, err_msg="observed e2 from EstimateShear changes when masking edge")
    np.testing.assert_almost_equal(ress.moments_sigma, ress_maskedge.moments_sigma,
        decimal=test_decimal,
        err_msg="observed sigma from EstimateShear changes when masking edge")
    np.testing.assert_almost_equal(ress.corrected_e1, ress_maskedge.corrected_e1,
        decimal=test_decimal,
        err_msg="corrected e1 from EstimateShear changes when masking edge")
    np.testing.assert_almost_equal(ress.corrected_e2, ress_maskedge.corrected_e2,
        decimal=test_decimal,
        err_msg="corrected e2 from EstimateShear changes when masking edge")
    np.testing.assert_almost_equal(ress.resolution_factor, ress_maskedge.resolution_factor,
        decimal=test_decimal,
        err_msg="resolution factor from EstimateShear changes when masking edge")

    # check that results don't change *at all* i.e. using assert_equal when we do this edge masking
    # in different ways:
    ## do the same as the previous test, but with weight map that is floats (0.0 or 1.0)
    maskedge = galsim.ImageF(imsize, imsize, init_value = 1.)
    for ind1 in range(xmin, xmax+1):
        for ind2 in range(ymin, ymax+1):
            if (ind1 <= (xmin+edgenum)) or (ind1 >= (xmax-edgenum)) or (ind2 <= (ymin+edgenum)) or (ind2 >= (ymax-edgenum)):
                maskedge.setValue(ind1, ind2, 0.)
    resm_maskedge1 = im.FindAdaptiveMom(maskedge)
    ress_maskedge1 = galsim.hsm.EstimateShear(im, p_im, maskedge)
    check_equal(resm_maskedge,ress_maskedge,resm_maskedge1,ress_maskedge1,
                "when masking with floats")

    ## make the weight map for allowed pixels a nonzero value that also != 1
    maskedge = galsim.ImageF(imsize, imsize, init_value = 2.3)
    for ind1 in range(xmin, xmax+1):
        for ind2 in range(ymin, ymax+1):
            if (ind1 <= (xmin+edgenum)) or (ind1 >= (xmax-edgenum)) or (ind2 <= (ymin+edgenum)) or (ind2 >= (ymax-edgenum)):
                maskedge.setValue(ind1, ind2, 0.)
    resm_maskedge1 = im.FindAdaptiveMom(maskedge)
    ress_maskedge1 = galsim.hsm.EstimateShear(im, p_im, maskedge)
    check_equal(resm_maskedge,ress_maskedge,resm_maskedge1,ress_maskedge1,
                "when masking with floats != 1")

    ## make the weight map all equal to 1, and use a badpix map with a range of nonzero values
    maskedge = galsim.ImageI(imsize, imsize, init_value = 1)
    badpixedge = galsim.ImageI(imsize, imsize, init_value = 0)
    for ind1 in range(xmin, xmax+1):
        for ind2 in range(ymin, ymax+1):
            if (ind1 <= (xmin+edgenum)) or (ind1 >= (xmax-edgenum)) or (ind2 <= (ymin+edgenum)) or (ind2 >= (ymax-edgenum)):
                badpixedge.setValue(ind1, ind2, ind1+1)
    resm_maskedge1 = im.FindAdaptiveMom(maskedge, badpixedge)
    ress_maskedge1 = galsim.hsm.EstimateShear(im, p_im, maskedge, badpixedge)
    check_equal(resm_maskedge,ress_maskedge,resm_maskedge1,ress_maskedge1,
                "when masking with badpix")

    ## same as previous, but with badpix of floats
    maskedge = galsim.ImageI(imsize, imsize, init_value = 1)
    badpixedge = galsim.ImageF(imsize, imsize, init_value = 0.)
    for ind1 in range(xmin, xmax+1):
        for ind2 in range(ymin, ymax+1):
            if (ind1 <= (xmin+edgenum)) or (ind1 >= (xmax-edgenum)) or (ind2 <= (ymin+edgenum)) or (ind2 >= (ymax-edgenum)):
                badpixedge.setValue(ind1, ind2, float(ind1+1))
    resm_maskedge1 = im.FindAdaptiveMom(maskedge, badpixedge)
    ress_maskedge1 = galsim.hsm.EstimateShear(im, p_im, maskedge, badpixedge)
    check_equal(resm_maskedge,ress_maskedge,resm_maskedge1,ress_maskedge1,
                "when masking with badpix (floats)")

    ## do some of the masking using weight map, and the rest using badpix
    maskedge = galsim.ImageI(imsize, imsize, init_value = 1)
    badpixedge = galsim.ImageI(imsize, imsize, init_value = 0)
    meanval = int(0.5*(xmin+xmax))
    for ind1 in range(xmin, xmax+1):
        for ind2 in range(ymin, ymax+1):
            if (ind1 <= (xmin+edgenum)) or (ind1 >= (xmax-edgenum)) or (ind2 <= (ymin+edgenum)) or (ind2 >= (ymax-edgenum)):
                if ind1 < meanval:
                    badpixedge.setValue(ind1, ind2, 1)
                else:
                    maskedge.setValue(ind1, ind2, 0)
    resm_maskedge1 = im.FindAdaptiveMom(maskedge, badpixedge)
    ress_maskedge1 = galsim.hsm.EstimateShear(im, p_im, maskedge, badpixedge)
    check_equal(resm_maskedge,ress_maskedge,resm_maskedge1,ress_maskedge1,
                "when masking with badpix and weight map")


@timer
def test_masks_with_check():
    """Test which errors in masks for moments and shear estimation routines are caught with check."""
    # Set up some toy galaxy and PSF
    my_sigma = 1.0
    my_pixscale = 0.1
    my_g1 = 0.15
    my_g2 = -0.4
    imsize = 256
    g = galsim.Gaussian(sigma = my_sigma)
    p = galsim.Gaussian(sigma = my_sigma)

    g = g.shear(g1=my_g1, g2=my_g2)
    obj = galsim.Convolve(g, p)
    im = galsim.ImageF(imsize, imsize)
    p_im = galsim.ImageF(imsize, imsize)
    obj.drawImage(image=im, scale=my_pixscale, method='no_pixel')
    p.drawImage(image=p_im, scale=my_pixscale, method='no_pixel')

    ## Create a weight image with size different from image
    ## Let it be smaller in one dimension, and larger in the other.
    bad_weight_im = galsim.ImageI(imsize, 2*imsize, init_value=1)
    small_weight_im = galsim.ImageI(2*imsize, imsize//4, init_value=1)
    # Setting strict=True (default) catches this error even if check=False.
    galsim.hsm.FindAdaptiveMom(im, small_weight_im, strict=False, check=False)
    assert_raises(galsim.errors.GalSimHSMError, galsim.hsm.FindAdaptiveMom, im, small_weight_im, strict=True, check=False)

    # A zero weight image raises errors,
    zero_weight_im = galsim.ImageI(imsize, 2*imsize, init_value=0)
    assert_raises(galsim.errors.GalSimHSMError, galsim.hsm.FindAdaptiveMom, im, zero_weight_im, check=False)
    assert_raises(galsim.errors.GalSimHSMError, galsim.hsm.EstimateShear, im, p_im, zero_weight_im, strict=True, check=False)
    # but a negative weight image can slip through with check=False.
    negative_weight_im = galsim.ImageI(imsize, 2*imsize, init_value=-1)
    galsim.hsm.FindAdaptiveMom(im, negative_weight_im, check=False)
    galsim.hsm.EstimateShear(im, p_im, negative_weight_im, check=False)
    # But these are unique to ImageI. Passing the same as ImageS will catch the error, but is more expensive due to deep copy.
    negative_weight_im_singleprecision = galsim.ImageS(imsize, imsize, init_value=-1)
    assert_raises(galsim.errors.GalSimHSMError, galsim.hsm.FindAdaptiveMom, im, negative_weight_im_singleprecision, check=False)

    # But all of these errors are caught with check=True.
    assert_raises(galsim.errors.GalSimIncompatibleValuesError, galsim.hsm.FindAdaptiveMom, im, bad_weight_im, check=True)
    assert_raises(galsim.errors.GalSimIncompatibleValuesError, galsim.hsm.FindAdaptiveMom, im, zero_weight_im, check=True)
    assert_raises(galsim.errors.GalSimIncompatibleValuesError, galsim.hsm.FindAdaptiveMom, im, negative_weight_im, check=True)
    assert_raises(galsim.errors.GalSimValueError, galsim.hsm.FindAdaptiveMom, im, negative_weight_im_singleprecision, check=True)
    assert_raises(galsim.errors.GalSimIncompatibleValuesError, galsim.hsm.EstimateShear, im, p_im, bad_weight_im, check=True)
    assert_raises(galsim.errors.GalSimIncompatibleValuesError, galsim.hsm.EstimateShear, im, p_im, zero_weight_im, check=True)
    assert_raises(galsim.errors.GalSimIncompatibleValuesError, galsim.hsm.EstimateShear, im, p_im, negative_weight_im, check=True)
    assert_raises(galsim.errors.GalSimValueError, galsim.hsm.EstimateShear, im, p_im, negative_weight_im_singleprecision, check=True)


@timer
def test_shearest_shape():
    """Test that shear estimation is insensitive to shape of input images."""
    # this test can help reveal bugs having to do with x / y indexing issues
    # just do test for one particular gaussian
    g1 = shear_values[1]
    g2 = shear_values[2]
    e1_psf = 0.05
    e2_psf = -0.04
    total_shear = np.sqrt(g1**2 + g2**2)
    conversion_factor = np.tanh(2.0*math.atanh(total_shear))/total_shear
    distortion_1 = g1*conversion_factor
    distortion_2 = g2*conversion_factor
    gal = galsim.Exponential(flux = 1.0, half_light_radius = 1.)
    gal = gal.shear(g1=g1, g2=g2)
    psf = galsim.Kolmogorov(flux = 1.0, fwhm = 0.7)
    psf = psf.shear(e1=e1_psf, e2=e2_psf)
    final = galsim.Convolve([gal, psf])

    imsize = [128, 256]
    for method_index in range(len(correction_methods)):
        print(correction_methods[method_index])

        save_e1 = -100.
        save_e2 = -100.
        for gal_x_imsize in imsize:
            for gal_y_imsize in imsize:
                for psf_x_imsize in imsize:
                    for psf_y_imsize in imsize:
                        final_image = galsim.ImageF(gal_x_imsize, gal_y_imsize)
                        epsf_image = galsim.ImageF(psf_x_imsize, psf_y_imsize)

                        final.drawImage(image=final_image, scale=pixel_scale, method='no_pixel')
                        psf.drawImage(image=epsf_image, scale=pixel_scale, method='no_pixel')
                        result = galsim.hsm.EstimateShear(final_image, epsf_image,
                            shear_est = correction_methods[method_index])
                        e1 = result.corrected_e1
                        e2 = result.corrected_e2
                        # make sure answers don't change as we vary image size

                        tot_e = np.sqrt(save_e1**2 + save_e2**2)
                        if tot_e < 99.:
                            np.testing.assert_almost_equal(e1, save_e1,
                                err_msg = "- incorrect e1",
                                decimal = decimal_shape)
                            np.testing.assert_almost_equal(e2, save_e2,
                                err_msg = "- incorrect e2",
                                decimal = decimal_shape)
                        save_e1 = e1
                        save_e2 = e2


@timer
def test_hsmparams():
    """Test the ability to set/change parameters that define how moments/shape estimation are done."""
    # First make some profile, and make sure that we get the same answers when we specify default
    # hsmparams or don't specify hsmparams at all.
    default_hsmparams = galsim.hsm.HSMParams(nsig_rg=3.0,
                                             nsig_rg2=3.6,
                                             regauss_too_small=1,
                                             adapt_order=2,
                                             convergence_threshold=1.e-6,
                                             max_mom2_iter=400,
                                             num_iter_default=-1,
                                             bound_correct_wt=0.25,
                                             max_amoment=8000.,
                                             max_ashift=15.,
                                             ksb_moments_max=4,
                                             failed_moments=-1000.)
    bulge = galsim.DeVaucouleurs(half_light_radius = 0.3)
    disk = galsim.Exponential(half_light_radius = 0.5)
    disk = disk.shear(e1=0.2, e2=-0.3)
    psf = galsim.Kolmogorov(fwhm = 0.6)
    gal = bulge + disk   # equal weighting, i.e., B/T=0.5
    tot_gal = galsim.Convolve(gal, psf)
    tot_gal_image = tot_gal.drawImage(scale=0.18)
    tot_psf_image = psf.drawImage(scale=0.18)

    res = tot_gal_image.FindAdaptiveMom()
    res_def = tot_gal_image.FindAdaptiveMom(hsmparams = default_hsmparams)
    assert(equal_hsmshapedata(res, res_def)), 'Moment outputs differ when using default HSMParams'
    assert res == res_def, 'Moment outputs differ when using default HSMParams'

    res2 = galsim.hsm.EstimateShear(tot_gal_image, tot_psf_image)
    res2_def = galsim.hsm.EstimateShear(tot_gal_image, tot_psf_image, hsmparams = default_hsmparams)
    assert(equal_hsmshapedata(res, res_def)), 'Shear outputs differ when using default HSMParams'
    assert res == res_def, 'Shear outputs differ when using default HSMParams'

    check_pickle(default_hsmparams)
    check_pickle(galsim.hsm.HSMParams(nsig_rg=1.0,
                                   nsig_rg2=1.6,
                                   regauss_too_small=0,
                                   adapt_order=0,
                                   convergence_threshold=1.e-8,
                                   max_mom2_iter=100,
                                   num_iter_default=4,
                                   bound_correct_wt=0.05,
                                   max_amoment=80.,
                                   max_ashift=5.,
                                   ksb_moments_max=2,
                                   failed_moments=99.))
    check_pickle(res)
    check_pickle(res2)

    # Then check failure modes: force it to fail by changing HSMParams.
    new_params_niter = galsim.hsm.HSMParams(max_mom2_iter = res.moments_n_iter-1)
    new_params_size = galsim.hsm.HSMParams(max_amoment = 0.3*res.moments_sigma**2)
    assert_raises(galsim.GalSimError, galsim.hsm.FindAdaptiveMom, tot_gal_image,
                  hsmparams=new_params_niter)
    assert_raises(galsim.GalSimError, galsim.hsm.EstimateShear, tot_gal_image, tot_psf_image,
                  hsmparams=new_params_size)

    assert_raises(TypeError, galsim.hsm.EstimateShear, tot_gal_image, tot_psf_image,
                  hsmparams='hsmparams')


@timer
def test_hsmparams_nodefault():
    """Test that when non-default hsmparams are used, the results change."""
    import time
    # First make some profile
    bulge = galsim.DeVaucouleurs(half_light_radius = 0.3)
    disk = galsim.Exponential(half_light_radius = 0.5)
    disk = disk.shear(e1=0.2, e2=-0.3)
    psf = galsim.Kolmogorov(fwhm = 0.6)
    gal = bulge + disk   # equal weighting, i.e., B/T=0.5
    tot_gal = galsim.Convolve(gal, psf)
    tot_gal_image = tot_gal.drawImage(scale=0.18)
    tot_psf_image = psf.drawImage(scale=0.18)

    # Check that recompute_flux changes give results that are as expected
    test_t = time.time()
    res = galsim.hsm.EstimateShear(tot_gal_image, tot_psf_image)
    dt = time.time() - test_t
    res2 = galsim.hsm.EstimateShear(tot_gal_image, tot_psf_image, recompute_flux = 'sum')
    assert(res.moments_amp < res2.moments_amp),'Incorrect behavior with recompute_flux=sum'
    res3 = galsim.hsm.EstimateShear(tot_gal_image, tot_psf_image, recompute_flux = 'none')
    assert(res3.moments_amp == 0),'Incorrect behavior with recompute_flux=none'

    # Check correction_status and error message when recompute_flux is invalid.
    with assert_raises(galsim.GalSimError):
        galsim.hsm.EstimateShear(tot_gal_image, tot_psf_image, recompute_flux='invalid')
    res4 = galsim.hsm.EstimateShear(tot_gal_image, tot_psf_image, recompute_flux='invalid',
                                    strict=False)
    assert res4.correction_status == -1
    assert "Unknown value" in res4.error_message

    # Check that results, timing change as expected with nsig_rg
    # For this, use Gaussian as galaxy and for ePSF, i.e., no extra pixel response
    p = galsim.Gaussian(fwhm=10.)
    g = galsim.Gaussian(fwhm=20.)
    g = g.shear(g1=0.5)
    obj = galsim.Convolve(g, p)
    # HSM allows a slop of 1.e-8 on nsig_rg, which means that default float32 images don't
    # actually end up with different result when using nsig_rg=0. rather than 3.
    im = obj.drawImage(scale=1., method='no_pixel', dtype=float)
    psf_im = p.drawImage(scale=1., method='no_pixel', dtype=float)
    test_t1 = time.time()
    g_res = galsim.hsm.EstimateShear(im, psf_im)
    test_t2 = time.time()
    g_res2 = galsim.hsm.EstimateShear(im, psf_im, hsmparams=galsim.hsm.HSMParams(nsig_rg=0.))
    dt2 = time.time()-test_t2
    dt1 = test_t2-test_t1
    if test_timing:
        assert(dt2 > dt1),'Should take longer to estimate shear without truncation of galaxy'
    assert(not equal_hsmshapedata(g_res, g_res2)),'Results should differ with diff nsig_rg'
    assert g_res != g_res2,'Results should differ with diff nsig_rg'

    # Check that results, timing change as expected with convergence_threshold
    test_t2 = time.time()
    res2 = galsim.hsm.EstimateShear(tot_gal_image, tot_psf_image,
                                    hsmparams=galsim.hsm.HSMParams(convergence_threshold = 1.e-3))
    dt2 = time.time() - test_t2
    if test_timing:
        assert(dt2 < dt),'Should be faster to estimate shear with higher convergence_threshold'
    assert(not equal_hsmshapedata(res, res2)),'Outputs same despite change in convergence_threshold'
    assert res != res2,'Outputs same despite change in convergence_threshold'

    # Check that max_amoment, max_ashift work as expected
    assert_raises(galsim.GalSimError,
        galsim.hsm.EstimateShear, tot_gal_image, tot_psf_image,
        hsmparams=galsim.hsm.HSMParams(max_amoment = 10.))
    assert_raises(galsim.GalSimError,
        galsim.hsm.EstimateShear, tot_gal_image, tot_psf_image,
        guess_centroid=galsim.PositionD(47., tot_gal_image.true_center.y),
        hsmparams=galsim.hsm.HSMParams(max_ashift=0.1))


@timer
def test_shapedata():
    """Check for basic issues with initialization of ShapeData objects."""
    x = 1.
    # Cannot initialize with messed up arguments.
    assert_raises(TypeError, galsim.hsm.ShapeData, x, x)
    assert_raises(TypeError, galsim.hsm.ShapeData, x)

    # Check that if initialized when empty, the resulting object has certain properties.
    foo = galsim.hsm.ShapeData()
    if foo.observed_shape != galsim.Shear() or foo.moments_n_iter != 0 or foo.meas_type != "None":
        raise AssertionError("Default ShapeData object was not as expected!")


@timer
def test_strict():
    """Check that using strict=True results in the behavior we expect."""
    # Set up an image for which moments measurement should fail spectacularly.
    scale = 2.7
    size = 11
    pix = galsim.Pixel(scale)
    image = galsim.Image(size, size)
    im = pix.drawImage(image=image, scale=scale, method='no_pixel')

    # Try to measure moments with strict = False. Make sure there's an error message stored.
    res = im.FindAdaptiveMom(strict = False)
    if res.error_message == '':
        raise AssertionError("Should have error message stored in case of FindAdaptiveMom failure!")

    # Check that measuring moments with strict = True results in the expected exception, and that
    # it is the same one as is stored when running with strict = False.
    with assert_raises(galsim.GalSimError):
        galsim.hsm.FindAdaptiveMom(im)
    try:
        res2 = im.FindAdaptiveMom()
    except galsim.GalSimError as err:
        if str(err) != res.error_message:
            raise AssertionError("Error messages do not match when running identical tests!")

    # Now redo the above for EstimateShear
    res = galsim.hsm.EstimateShear(im, im, strict = False)
    if res.error_message == '':
        raise AssertionError("Should have error message stored in case of EstimateShear failure!")
    with assert_raises(galsim.GalSimError):
        galsim.hsm.EstimateShear(im, im)
    try:
        res2 = galsim.hsm.EstimateShear(im, im)
    except galsim.GalSimError as err:
        if str(err) != res.error_message:
            raise AssertionError("Error messages do not match when running identical tests!")


@timer
def test_bounds_centroid():
    """Check that the input bounds are respected, and centroid coordinates make sense."""
    # Make a simple object drawn into image with non-trivial bounds (even-sized).
    b = galsim.BoundsI(37, 326, 47, 336)
    test_scale = 0.15
    test_sigma = 3.1
    im = galsim.Image(bounds=b)
    im.scale = test_scale

    obj = galsim.Gaussian(sigma=test_sigma)
    obj.drawImage(image=im, scale=test_scale, method='no_pixel')
    mom = im.FindAdaptiveMom()
    np.testing.assert_almost_equal(
        mom.moments_centroid.x, im.true_center.x, decimal=7,
        err_msg='Moments x centroid differs from true center of even-sized image')
    np.testing.assert_almost_equal(
        mom.moments_centroid.y, im.true_center.y, decimal=7,
        err_msg='Moments y centroid differs from true center of even-sized image')

    # Draw the same object into odd-sized image with non-trivial bounds.
    b2 = galsim.BoundsI(b.xmin, b.xmax+1, b.ymin, b.ymax+1)
    im = galsim.Image(bounds=b2)
    im.scale = test_scale
    obj.drawImage(image=im, scale=test_scale, method='no_pixel')
    mom = im.FindAdaptiveMom()
    np.testing.assert_almost_equal(
        mom.moments_centroid.x, im.true_center.x, decimal=7,
        err_msg='Moments x centroid differs from true center of odd-sized image')
    np.testing.assert_almost_equal(
        mom.moments_centroid.y, im.true_center.y, decimal=7,
        err_msg='Moments y centroid differs from true center of odd-sized image')

    # Check that it still works with a symmetric sub-image.
    sub_im = im[galsim.BoundsI(b2.xmin+2, b2.xmax-2, b2.ymin+2, b2.ymax-2)]
    mom = sub_im.FindAdaptiveMom()
    np.testing.assert_almost_equal(
        mom.moments_centroid.x, sub_im.true_center.x, decimal=7,
        err_msg='Moments x centroid differs from true center of odd-sized subimage')
    np.testing.assert_almost_equal(
        mom.moments_centroid.y, sub_im.true_center.y, decimal=7,
        err_msg='Moments y centroid differs from true center of odd-sized subimage')

    # Check that we can take a weird/asymmetric sub-image, and it fails because of centroid shift.
    sub_im = im[galsim.BoundsI(b2.xmin, b2.xmax-100, b2.ymin+27, b2.ymax)]
    with assert_raises(galsim.GalSimError):
        galsim.hsm.FindAdaptiveMom(sub_im)

    # ... and that it passes if we hand in a good centroid guess.  Note that this test is a bit less
    # stringent than some of the previous ones, because our subimage cut off a decent part of the
    # light profile in the x direction, affecting the x centroid estimate.  But the y centroid test
    # is the same precision as before.
    mom = sub_im.FindAdaptiveMom(guess_centroid=im.true_center)
    np.testing.assert_approx_equal(
        mom.moments_centroid.x, im.true_center.x, significant=4,
        err_msg='Moments x centroid differs from true center of asymmetric subimage')
    np.testing.assert_almost_equal(
        mom.moments_centroid.y, im.true_center.y, decimal=7,
        err_msg='Moments y centroid differs from true center of asymmetric subimage')


@timer
def test_ksb_sig():
    """Check that modification of KSB weight function width works."""
    gal = galsim.Gaussian(fwhm=1.0).shear(e1=0.2, e2=0.1)
    psf = galsim.Gaussian(fwhm=0.7)
    gal_img = galsim.Convolve(gal, psf).drawImage(nx=32, ny=32, scale=0.2)
    psf_img = psf.drawImage(nx=16, ny=16, scale=0.2)

    # First just check that combination of ksb_sig_weight and ksb_sig_factor is consistent.
    hsmparams1 = galsim.hsm.HSMParams(ksb_sig_weight=2.0)
    result1 = galsim.hsm.EstimateShear(gal_img, psf_img, shear_est='KSB', hsmparams=hsmparams1)

    hsmparams2 = galsim.hsm.HSMParams(ksb_sig_weight=1.0, ksb_sig_factor=2.0)
    result2 = galsim.hsm.EstimateShear(gal_img, psf_img, shear_est='KSB', hsmparams=hsmparams2)

    np.testing.assert_almost_equal(result1.corrected_g1, result2.corrected_g1, 9,
                                   "KSB weight fn width inconsistently manipulated")
    np.testing.assert_almost_equal(result1.corrected_g2, result2.corrected_g2, 9,
                                   "KSB weight fn width inconsistently manipulated")

    # Now check that if we construct a galaxy with an ellipticity gradient, we see the appropriate
    # sign of the response when we change the width of the weight function.
    narrow = galsim.Gaussian(fwhm=1.0).shear(e1=0.2)
    wide = galsim.Gaussian(fwhm=2.0).shear(e1=-0.2)
    gal = narrow + wide
    gal_img = galsim.Convolve(gal, psf).drawImage(nx=32, ny=32, scale=0.2)
    hsmparams_narrow = galsim.hsm.HSMParams()  # Default sig_factor=1.0
    result_narrow = galsim.hsm.EstimateShear(gal_img, psf_img, shear_est='KSB',
                                             hsmparams=hsmparams_narrow)
    hsmparams_wide = galsim.hsm.HSMParams(ksb_sig_factor=2.0)
    result_wide = galsim.hsm.EstimateShear(gal_img, psf_img, shear_est='KSB',
                                           hsmparams=hsmparams_wide)

    np.testing.assert_array_less(result_wide.corrected_g1, result_narrow.corrected_g1,
                                 "Galaxy ellipticity gradient not captured by ksb_sig_factor.")

@timer
def test_noncontiguous():
    """Test running HSM module with non-C-contiguous images.

    This test was inspired by a bug report in issue #833.
    """
    gal = galsim.Gaussian(sigma=1.3).shear(g1=0.1, g2=0.2)

    # First a normal C-contiguous image as build by drawImage.
    img = gal.drawImage(nx=64, ny=64, scale=0.2)
    meas_shape1 = galsim.hsm.FindAdaptiveMom(img).observed_shape
    print(meas_shape1)
    np.testing.assert_almost_equal(meas_shape1.g1, 0.1, decimal=3,
                                   err_msg="HSM measured wrong shear on normal image")
    np.testing.assert_almost_equal(meas_shape1.g2, 0.2, decimal=3,
                                   err_msg="HSM measured wrong shear on normal image")

    # Transpose the image, which should just flip the sign of g1.
    # Note, though, that this changes the ndarray from C-ordering to FORTRAN-ordering.
    fimg = galsim.Image(img.array.T, scale=0.2)
    meas_shape2 = galsim.hsm.FindAdaptiveMom(fimg).observed_shape
    print(meas_shape2)
    np.testing.assert_almost_equal(meas_shape2.g1, -0.1, decimal=3,
                                   err_msg="HSM measured wrong shear on transposed image")
    np.testing.assert_almost_equal(meas_shape2.g2, 0.2, decimal=3,
                                   err_msg="HSM measured wrong shear on transposed image")

    # Also test the real part of an ImageC, which not contiguous in either direction (step=2)
    # This should have the negative shear from drawing in k space.
    kimg = gal.drawKImage(nx=64, ny=64)
    meas_shape3 = galsim.hsm.FindAdaptiveMom(kimg.real).observed_shape
    print(meas_shape3)
    np.testing.assert_almost_equal(meas_shape3.g1, -0.1, decimal=3,
                                   err_msg="HSM measured wrong shear on image with step=2")
    np.testing.assert_almost_equal(meas_shape3.g2, -0.2, decimal=3,
                                   err_msg="HSM measured wrong shear on image with step=2")

@timer
def test_headers():
    # This isn't really an HSM test per se, but it's testing a feature that we added so
    # LSST DM can use the C++-layer HSM code from their C++ code.
    # It also adds a bit of stability guarantee to the HSM function signatures at least.

    # Check that the named files exist.
    include_dir = galsim.include_dir
    lib_file = galsim.lib_file
    print('include_dir = ',include_dir)
    print('lib_file = ',lib_file)
    assert os.path.isfile(lib_file)
    assert os.path.isfile(os.path.join(include_dir, 'GalSim.h'))

    # Check version code
    version_h = os.path.join(include_dir, 'galsim', 'Version.h')
    with open(version_h) as fin:
        lines = fin.read()
    assert '#define GALSIM_MAJOR %d'%(galsim.__version_info__[0]) in lines
    assert '#define GALSIM_MINOR %d'%(galsim.__version_info__[1]) in lines

    # Check the function signatudes of the code that LSST DM uses from the C++ layer.
    psfcorr_h = os.path.join(include_dir, 'galsim', 'hsm', 'PSFCorr.h')
    assert os.path.isfile(os.path.join(include_dir, 'galsim', 'hsm', 'PSFCorr.h'))

    fam_signature = """
    PUBLIC_API void FindAdaptiveMomView(
        ShapeData& results,
        const BaseImage<T> &object_image, const BaseImage<int> &object_mask_image,
        double guess_sig = 5.0, double precision = 1.0e-6,
        galsim::Position<double> guess_centroid = galsim::Position<double>(-1000.,-1000.),
        bool round_moments = false,
        const HSMParams& hsmparams=HSMParams());"""

    esv_signature = """
    PUBLIC_API void EstimateShearView(
        ShapeData& results,
        const BaseImage<T> &gal_image, const BaseImage<U> &PSF_image,
        const BaseImage<int> &gal_mask_image,
        float sky_var = 0.0, const char *shear_est = "REGAUSS",
        const char* recompute_flux = "FIT",
        double guess_sig_gal = 5.0, double guess_sig_PSF = 3.0, double precision = 1.0e-6,
        galsim::Position<double> guess_centroid = galsim::Position<double>(-1000.,-1000.),
        const HSMParams& hsmparams=HSMParams());"""

    # Note: If these fail (because we change the above signature for some reason), let LSST-DM
    # devs know that their code will break on the next version of GalSim.
    # They know we don't guarantee this to be stable, but it's a courtesy to let them know if
    # they might have problems.
    with open(psfcorr_h) as fin:
        lines = fin.read()
    assert fam_signature in lines
    assert esv_signature in lines

    # Check that the library is loadable.  (C++ name mangling means we can't actually check
    # that the right functions are there, but that should be implicitly tested by the fact
    # that calls from python actaully work...)
    import ctypes
    lib = ctypes.cdll.LoadLibrary(galsim.lib_file)
    # The test was that this doesn't raise an OSError or something.

@timer
def test_failures():
    """Test some images that used to fail, but now work.
    """
    files = glob.glob(os.path.join(img_dir, 'HSM*.fits'))

    for f in files:
        im = galsim.fits.read(f)
        hsm = im.FindAdaptiveMom()
        assert hsm.moments_status == 0

@timer
def test_very_small():
    """Test an unresolved star reported to fail in #1132, but now works.
    """
    profile = galsim.VonKarman(lam=700.0, r0=0.25, L0=11.0, flux=1.0).shift(-0.13, -0.13)
    im = profile.drawImage(nx=31, ny=31, scale=0.26)
    mom = im.FindAdaptiveMom(strict=False)
    print(mom)
    assert mom.moments_status == 0

    # However, if the object is small enough, then HSM cannot recover.  Make sure the
    # error message in such cases is sensible.
    profile = galsim.Gaussian(sigma=0.18).shift(0.02,-0.03)
    im = profile.drawImage(nx=31, ny=31, scale=0.26)
    mom = im.FindAdaptiveMom(strict=False)
    print(mom)
    assert mom.moments_status != 0
    assert "Object is too small" in mom.error_message


@timer
def test_negative_stepstride():
    """In response to #1185, check that hsm works for arrays with negative step or stride.
    """
    img = galsim.Gaussian(fwhm=1).drawImage()
    result1 = galsim.hsm.FindAdaptiveMom(img)
    result2 = galsim.hsm.FindAdaptiveMom(galsim.Image(img.array[::-1]))
    result3 = galsim.hsm.FindAdaptiveMom(galsim.Image(img.array[:,::-1]))
    result4 = galsim.hsm.FindAdaptiveMom(galsim.Image(img.array[::-1,::-1]))

    assert np.isclose(result1.moments_sigma, result2.moments_sigma)
    assert np.isclose(result1.moments_sigma, result3.moments_sigma)
    assert np.isclose(result1.moments_sigma, result4.moments_sigma)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Sersic', 'DeVaucouleurs' ]

import math

from . import _galsim
from .gsobject import GSObject
from .gsparams import GSParams
from .utilities import lazy_property, doc_inherit
from .errors import GalSimRangeError, GalSimIncompatibleValuesError, convert_cpp_errors

class Sersic(GSObject):
    r"""A class describing a Sersic profile.

    The Sersic surface brightness profile is characterized by three properties: its Sersic index
    ``n``, its ``flux``, and either the ``half_light_radius`` or ``scale_radius``.  Given these
    properties, the surface brightness profile scales as

    .. math::
        I(r) \sim e^{-(r/r_0)^{1/n}}

    where :math:`r_0` is the ``scale_radius``, or

    .. math::
        I(r) \sim e^{-b (r/r_e)^{1/n}}

    where :math:`r_e` is the ``half_light_radius`` and :math:`b` is calculated to give the right
    half-light radius.

    For more information, refer to

    http://en.wikipedia.org/wiki/Sersic_profile

    The allowed range of values for the ``n`` parameter is 0.3 <= n <= 6.2.  An exception will be
    thrown if you provide a value outside that range.  Below n=0.3, there are severe numerical
    problems.  Above n=6.2, we found that the code begins to be inaccurate when sheared or
    magnified (at the level of upcoming shear surveys), so we do not recommend extending beyond
    this.  See Issues #325 and #450 for more details.

    Sersic profile calculations take advantage of Hankel transform tables that are precomputed for a
    given value of n when the Sersic profile is initialized.  Making additional objects with the
    same n can therefore be many times faster than making objects with different values of n that
    have not been used before.  Moreover, these Hankel transforms are only cached for a maximum of
    100 different n values at a time.  For this reason, for large sets of simulations, it is worth
    considering the use of only discrete n values rather than allowing it to vary continuously.  For
    more details, see https://github.com/GalSim-developers/GalSim/issues/566.

    Note that if you are building many Sersic profiles using truncation, the code will be more
    efficient if the truncation is always the same multiple of ``scale_radius``, since it caches
    many calculations that depend on the ratio ``trunc/scale_radius``.

    A Sersic can be initialized using one (and only one) of two possible size parameters:
    ``scale_radius`` or ``half_light_radius``.  Exactly one of these two is required.

    Flux of a truncated profile:

    If you are truncating the profile, the optional parameter, ``flux_untruncated``, specifies
    whether the ``flux`` and ``half_light_radius`` specifications correspond to the untruncated
    profile (``True``) or to the truncated profile (``False``, default).  The impact of this
    parameter is a little subtle, so we'll go through a few examples to show how it works.

    First, let's examine the case where we specify the size according to the half-light radius.
    If ``flux_untruncated`` is True (and ``trunc > 0``), then the profile will be identical
    to the version without truncation up to the truncation radius, beyond which it drops to 0.
    In this case, the actual half-light radius will be different from the specified half-light
    radius.  The half_light_radius property will return the true half-light radius.  Similarly,
    the actual flux will not be the same as the specified value; the true flux is also returned
    by the flux property.

    Example::

        >>> sersic_obj1 = galsim.Sersic(n=3.5, half_light_radius=2.5, flux=40.)
        >>> sersic_obj2 = galsim.Sersic(n=3.5, half_light_radius=2.5, flux=40., trunc=10.)
        >>> sersic_obj3 = galsim.Sersic(n=3.5, half_light_radius=2.5, flux=40., trunc=10., \\
                                        flux_untruncated=True)

        >>> sersic_obj1.xValue(galsim.PositionD(0.,0.))
        237.3094228615618
        >>> sersic_obj2.xValue(galsim.PositionD(0.,0.))
        142.54505376530574    # Normalization and scale radius adjusted (same half-light radius)
        >>> sersic_obj3.xValue(galsim.PositionD(0.,0.))
        237.30942286156187

        >>> sersic_obj1.xValue(galsim.PositionD(10.0001,0.))
        0.011776164687304694
        >>> sersic_obj2.xValue(galsim.PositionD(10.0001,0.))
        0.0
        >>> sersic_obj3.xValue(galsim.PositionD(10.0001,0.))
        0.0

        >>> sersic_obj1.half_light_radius
        2.5
        >>> sersic_obj2.half_light_radius
        2.5
        >>> sersic_obj3.half_light_radius
        1.9795101383056892    # The true half-light radius is smaller than the specified value

        >>> sersic_obj1.flux
        40.0
        >>> sersic_obj2.flux
        40.0
        >>> sersic_obj3.flux
        34.56595186009519     # Flux is missing due to truncation

        >>> sersic_obj1.scale_radius
        0.003262738739834598
        >>> sersic_obj2.scale_radius
        0.004754602453641744  # the scale radius needed adjustment to accommodate HLR
        >>> sersic_obj3.scale_radius
        0.003262738739834598  # the scale radius is still identical to the untruncated case

    When the truncated Sersic scale is specified with ``scale_radius``, the behavior between the
    three cases (untruncated, ``flux_untruncated=True`` and ``flux_untruncated=False``) will be
    somewhat different from above.  Since it is the scale radius that is being specified, and since
    truncation does not change the scale radius the way it can change the half-light radius, the
    scale radius will remain unchanged in all cases.  This also results in the half-light radius
    being the same between the two truncated cases (although different from the untruncated case).
    The flux normalization is the only difference between ``flux_untruncated=True`` and
    ``flux_untruncated=False`` in this case.

    Example::

        >>> sersic_obj1 = galsim.Sersic(n=3.5, scale_radius=0.05, flux=40.)
        >>> sersic_obj2 = galsim.Sersic(n=3.5, scale_radius=0.05, flux=40., trunc=10.)
        >>> sersic_obj3 = galsim.Sersic(n=3.5, scale_radius=0.05, flux=40., trunc=10., \\
                                        flux_untruncated=True)

        >>> sersic_obj1.xValue(galsim.PositionD(0.,0.))
        1.010507575186637
        >>> sersic_obj2.xValue(galsim.PositionD(0.,0.))
        5.786692612210923     # Normalization adjusted to accomodate the flux within trunc radius
        >>> sersic_obj3.xValue(galsim.PositionD(0.,0.))
        1.010507575186637

        >>> sersic_obj1.half_light_radius
        38.311372735390016
        >>> sersic_obj2.half_light_radius
        5.160062547614234
        >>> sersic_obj3.half_light_radius
        5.160062547614234     # For the truncated cases, the half-light radii are the same

        >>> sersic_obj1.flux
        40.0
        >>> sersic_obj2.flux
        40.0
        >>> sersic_obj3.flux
        6.985044085834393     # Flux is missing due to truncation

        >>> sersic_obj1.scale_radius
        0.05
        >>> sersic_obj2.scale_radius
        0.05
        >>> sersic_obj3.scale_radius
        0.05
        half_light_radius:  The half-light radius

    Parameters:
        n:                  The Sersic index, n.
        half_light_radius:  The half-light radius of the profile.  Typically given in arcsec.
                            [One of ``scale_radius`` or ``half_light_radius`` is required.]
        scale_radius:       The scale radius of the profile.  Typically given in arcsec.
                            [One of ``scale_radius`` or ``half_light_radius`` is required.]
        flux:               The flux (in photons/cm^2/s) of the profile. [default: 1]
        trunc:              An optional truncation radius at which the profile is made to drop to
                            zero, in the same units as the size parameter.
                            [default: 0, indicating no truncation]
        flux_untruncated:   Should the provided ``flux`` and ``half_light_radius`` refer to the
                            untruncated profile? See below for more details. [default: False]
        gsparams:           An optional `GSParams` argument. [default: None]
    """
    _req_params = { "n" : float }
    _opt_params = { "flux" : float, "trunc" : float, "flux_untruncated" : bool }
    _single_params = [ { "scale_radius" : float , "half_light_radius" : float } ]

    _is_axisymmetric = True
    _is_analytic_x = True
    _is_analytic_k = True

    _minimum_n = 0.3  # Lower bounds has hard limit at ~0.29
    _maximum_n = 6.2  # Upper bounds is just where we have tested that code works well.

    # The conversion from hlr to scale radius is complicated for Sersic, especially since we
    # allow it to be truncated.  So we do these calculations in the C++-layer constructor.
    def __init__(self, n, half_light_radius=None, scale_radius=None,
                 flux=1., trunc=0., flux_untruncated=False, gsparams=None):
        self._n = float(n)
        self._flux = float(flux)
        self._trunc = float(trunc)
        self._gsparams = GSParams.check(gsparams)

        if self._n < Sersic._minimum_n:
            raise GalSimRangeError("Requested Sersic index is too small",
                                   self._n, Sersic._minimum_n, Sersic._maximum_n)
        if self._n > Sersic._maximum_n:
            raise GalSimRangeError("Requested Sersic index is too large",
                                   self._n, Sersic._minimum_n, Sersic._maximum_n)

        if self._trunc < 0:
            raise GalSimRangeError("Sersic trunc must be > 0", self._trunc, 0.)

        # Parse the radius options
        if half_light_radius is not None:
            if scale_radius is not None:
                raise GalSimIncompatibleValuesError(
                    "Only one of scale_radius or half_light_radius may be specified for Spergel",
                    half_light_radius=half_light_radius, scale_radius=scale_radius)
            self._hlr = float(half_light_radius)
            if self._trunc == 0. or flux_untruncated:
                self._flux_fraction = 1.
                self._r0 = self._hlr / self.calculateHLRFactor()
            else:
                if self._trunc <= math.sqrt(2.) * self._hlr:
                    raise GalSimRangeError("Sersic trunc must be > sqrt(2) * half_light_radius",
                                           self._trunc, math.sqrt(2.) * self._hlr)
                with convert_cpp_errors():
                    self._r0 = _galsim.SersicTruncatedScale(self._n, self._hlr, self._trunc)
        elif scale_radius is not None:
            self._r0 = float(scale_radius)
            self._hlr = 0.
        else:
            raise GalSimIncompatibleValuesError(
                "Either scale_radius or half_light_radius must be specified for Spergel",
                half_light_radius=half_light_radius, scale_radius=scale_radius)

        if self._trunc > 0.:
            self._flux_fraction = self.calculateIntegratedFlux(self._trunc)
            if flux_untruncated:
                # Then update the flux and hlr with the correct values
                self._flux *= self._flux_fraction
                self._hlr = 0.  # This will be updated by getHalfLightRadius if necessary.
        else:
            self._flux_fraction = 1.

    def calculateIntegratedFlux(self, r):
        """Return the fraction of the total flux enclosed within a given radius, r"""
        with convert_cpp_errors():
            return _galsim.SersicIntegratedFlux(self._n, float(r)/self._r0)

    def calculateHLRFactor(self):
        """Calculate the half-light-radius in units of the scale radius.
        """
        with convert_cpp_errors():
            return _galsim.SersicHLR(self._n, self._flux_fraction)

    @lazy_property
    def _sbp(self):
        with convert_cpp_errors():
            return _galsim.SBSersic(self._n, self._r0, self._flux, self._trunc, self.gsparams._gsp)

    @property
    def n(self):
        """The Sersic parameter n.
        """
        return self._n

    @property
    def scale_radius(self):
        """The scale radius.
        """
        return self._r0

    @property
    def trunc(self):
        """The truncation radius (if any).
        """
        return self._trunc

    @property
    def half_light_radius(self):
        """The half-light radius.
        """
        if self._hlr == 0.:
            self._hlr = self._r0 * self.calculateHLRFactor()
        return self._hlr

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, Sersic) and
                 self.n == other.n and
                 self.scale_radius == other.scale_radius and
                 self.trunc == other.trunc and
                 self.flux == other.flux and
                 self.gsparams == other.gsparams))

    def __hash__(self):
        return hash(("galsim.SBSersic", self.n, self.scale_radius, self.trunc, self.flux,
                     self.gsparams))

    def __repr__(self):
        return 'galsim.Sersic(n=%r, scale_radius=%r, trunc=%r, flux=%r, gsparams=%r)'%(
            self.n, self.scale_radius, self.trunc, self.flux, self.gsparams)

    def __str__(self):
        # Note: for the repr, we use the scale_radius, since that should just flow as is through
        # the constructor, so it should be exact.  But most people use half_light_radius
        # for Sersics, so use that in the looser str() function.
        s = 'galsim.Sersic(n=%s, half_light_radius=%s'%(self.n, self.half_light_radius)
        if self.trunc != 0.:
            s += ', trunc=%s'%self.trunc
        if self.flux != 1.0:
            s += ', flux=%s'%self.flux
        s += ')'
        return s

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    @property
    def _maxk(self):
        return self._sbp.maxK()

    @property
    def _stepk(self):
        return self._sbp.stepK()

    @property
    def _has_hard_edges(self):
        return self._trunc != 0.

    @property
    def _max_sb(self):
        return self._sbp.maxSB()

    def _xValue(self, pos):
        return self._sbp.xValue(pos._p)

    def _kValue(self, kpos):
        return self._sbp.kValue(kpos._p)

    def _drawReal(self, image, jac=None, offset=(0.,0.), flux_scaling=1.):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        dx,dy = offset
        self._sbp.draw(image._image, image.scale, _jac, dx, dy, flux_scaling)

    def _shoot(self, photons, rng):
        self._sbp.shoot(photons._pa, rng._rng)

    def _drawKImage(self, image, jac=None):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        self._sbp.drawK(image._image, image.scale, _jac)

    @doc_inherit
    def withFlux(self, flux):
        return Sersic(n=self.n, scale_radius=self.scale_radius, trunc=self.trunc, flux=flux,
                      gsparams=self.gsparams)

class DeVaucouleurs(Sersic):
    r"""A class describing DeVaucouleurs profile objects.

    Surface brightness profile with

    .. math::
        I(r) \sim e^{-(r/r_0)^{1/4}}

    where :math:`r_0` is the ``scale_radius``. This is completely equivalent to a Sersic with n=4.

    For more information, refer to

    http://en.wikipedia.org/wiki/De_Vaucouleurs'_law

    A DeVaucouleurs can be initialized using one (and only one) of two possible size parameters:
    ``scale_radius`` or ``half_light_radius``.  Exactly one of these two is required.

    Parameters:
        scale_radius:       The value of scale radius of the profile.  Typically given in arcsec.
                            [One of ``scale_radius`` or ``half_light_radius`` is required.]
        half_light_radius:  The half-light radius of the profile.  Typically given in arcsec.
                            [One of ``scale_radius`` or ``half_light_radius`` is required.]
        flux:               The flux (in photons/cm^2/s) of the profile. [default: 1]
        trunc:              An optional truncation radius at which the profile is made to drop to
                            zero, in the same units as the size parameter.
                            [default: 0, indicating no truncation]
        flux_untruncated:   Should the provided ``flux`` and ``half_light_radius`` refer to the
                            untruncated profile? See the docstring for Sersic for more details.
                            [default: False]
        gsparams:           An optional `GSParams` argument. [default: None]
    """
    _req_params = {}
    _opt_params = { "flux" : float, "trunc" : float, "flux_untruncated" : bool }
    _single_params = [ { "scale_radius" : float , "half_light_radius" : float } ]

    def __init__(self, half_light_radius=None, scale_radius=None, flux=1., trunc=0.,
                 flux_untruncated=False, gsparams=None):
        super(DeVaucouleurs, self).__init__(n=4, half_light_radius=half_light_radius,
                                            scale_radius=scale_radius, flux=flux,
                                            trunc=trunc, flux_untruncated=flux_untruncated,
                                            gsparams=gsparams)

    def __repr__(self):
        return 'galsim.DeVaucouleurs(scale_radius=%r, trunc=%r, flux=%r, gsparams=%r)'%(
            self.scale_radius, self.trunc, self.flux, self.gsparams)

    def __str__(self):
        s = 'galsim.DeVaucouleurs(half_light_radius=%s'%self.half_light_radius
        if self.trunc != 0.:
            s += ', trunc=%s'%self.trunc
        if self.flux != 1.0:
            s += ', flux=%s'%self.flux
        s += ')'
        return s

    @doc_inherit
    def withFlux(self, flux):
        return DeVaucouleurs(scale_radius=self.scale_radius, trunc=self.trunc, flux=flux,
                             gsparams=self.gsparams)
